# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from typing import Any

import numpy as np
import torch
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from torchcp.utils.registry import Registry

METRICS_REGISTRY_CLASSIFICATION = Registry("METRICS")


#########################################
# Marginal coverage metric
#########################################

@METRICS_REGISTRY_CLASSIFICATION.register()
def coverage_rate(prediction_sets, labels, coverage_type="default", num_classes=None):
    """
    The metric for empirical coverage.
    
    Args:
        prediction_sets (torch.Tensor): Boolean tensor of prediction sets (N x C), where N is number of samples
            and C is number of classes.
        labels (torch.Tensor): Ground-truth labels (N,).
        coverage_type (str, optional): Type of coverage rate calculation.
            'default': marginal coverage rate
            'macro': average coverage rate across all classes
        num_classes (int, optional): Number of classes. Required when coverage_type is 'macro'.
    Returns:
        float: Empirical coverage rate.
    """
    labels = labels.cpu()
    prediction_sets = prediction_sets.cpu()

    if prediction_sets.shape[0] != len(labels):
        raise ValueError("The number of prediction sets must be equal to the number of labels.")

    if coverage_type not in ["default", "macro"]:
        raise ValueError("coverage_type must be 'default' or 'macro'.")

    cvg = 0

    covered = prediction_sets[torch.arange(len(labels)), labels]

    if coverage_type == "macro":
        if num_classes is None:
            raise ValueError("When coverage_type is 'macro', you must define the number of classes.")

        class_counts = torch.bincount(labels, minlength=num_classes)
        class_covered = torch.bincount(labels[covered == 1], minlength=num_classes)

        class_coverage = torch.zeros_like(class_counts, dtype=torch.float)
        valid_classes = class_counts > 0
        class_coverage[valid_classes] = class_covered[valid_classes].float() / class_counts[valid_classes].float()

        cvg = class_coverage.mean().item()
    else:
        cvg = covered.float().mean().item()
    return cvg


@METRICS_REGISTRY_CLASSIFICATION.register()
def average_size(prediction_sets, labels=None):
    return torch.mean(torch.sum(prediction_sets, dim=1).float()).item()


#########################################
# Conditional coverage metric
#########################################

@METRICS_REGISTRY_CLASSIFICATION.register()
def CovGap(prediction_sets, labels, alpha, num_classes, shot_idx=None):
    """
    The average class-conditional coverage gap.

    Paper: Class-Conditional Conformal Prediction with Many Classes (Ding et al., 2023)
    Link: https://neurips.cc/virtual/2023/poster/70548
    
    Args:
        prediction_sets (torch.Tensor): Boolean tensor of prediction sets (N x C).
        labels (torch.Tensor): Ground-truth labels (N,).
        alpha (float): User-guided confidence level.
        num_classes (int): Number of classes.
        shot_idx (list, optional): Indices of classes to compute coverage gap.
    
    Returns:
        float: Average class-conditional coverage gap (percentage).
    """
    if prediction_sets.shape[0] != len(labels):
        raise ValueError("Number of prediction sets must match number of labels")

    labels = labels.cpu()
    prediction_sets = prediction_sets.cpu()

    covered = prediction_sets[torch.arange(len(labels)), labels]
    class_counts = torch.bincount(labels, minlength=num_classes)
    class_covered = torch.bincount(labels[covered == 1], minlength=num_classes)

    cls_coverage_rate = torch.zeros_like(class_counts, dtype=torch.float32)
    valid_classes = class_counts > 0
    cls_coverage_rate[valid_classes] = class_covered[valid_classes].float() / class_counts[valid_classes].float()

    if shot_idx is not None:
        mask = torch.zeros(num_classes, dtype=torch.bool)
        mask[shot_idx] = True
        valid_classes = valid_classes & mask
        
    overall_covgap = torch.mean(torch.abs(cls_coverage_rate[valid_classes] - (1 - alpha))) * 100
    return overall_covgap.float().item()


@METRICS_REGISTRY_CLASSIFICATION.register()
def VioClasses(prediction_sets, labels, alpha, num_classes):
    """
    The number of violated classes.

    Paper: Empirically Validating Conformal Prediction on Modern Vision Architectures Under Distribution Shift and Long-tailed Data (Kasa et al., 2023)
    Link: https://arxiv.org/abs/2307.01088
    
    Args:
        prediction_sets (torch.Tensor): Boolean tensor of prediction sets (N x C).
        labels (torch.Tensor): Ground-truth labels (N,).
        alpha (float): User-guided confidence level.
        num_classes (int): Number of classes.
    
    Returns:
        int: Number of classes with violated coverage.
    """
    if prediction_sets.shape[0] != len(labels):
        raise ValueError("Number of prediction sets must match number of labels")

    labels = labels.cpu()
    prediction_sets = prediction_sets.cpu()
    violation_nums = 0

    covered = prediction_sets[torch.arange(len(labels)), labels]

    class_covered = torch.bincount(labels[covered == 1], minlength=num_classes)
    class_counts = torch.bincount(labels, minlength=num_classes)

    valid_classes = class_counts > 0
    class_coverage = torch.zeros(num_classes, dtype=torch.float32)
    class_coverage[valid_classes] = class_covered[valid_classes].float() / class_counts[valid_classes].float()
    
    violation_nums = torch.sum(class_coverage[valid_classes] < (1 - alpha))
    return violation_nums.item()


@METRICS_REGISTRY_CLASSIFICATION.register()
def DiffViolation(logits, prediction_sets, labels, alpha,
                  strata_diff=[[1, 1], [2, 3], [4, 6], [7, 10], [11, 100], [101, 1000]]):
    """
    Difficulty-stratified coverage violation

    Paper: Uncertainty Sets for Image Classifiers using Conformal Prediction (Angelopoulos et al., 2020)
    Link: https://arxiv.org/abs/2009.14193
    
    
    Args:
        logits (torch.Tensor): the predicted logits.
        prediction_sets (torch.Tensor): the prediction sets generated by CP algorithms.
        labels (list): the ground-truth label of each samples.
        alpha (float): the user-guided confidence level.
        strata_diff (list): a coarse partitioning of the possible difficulties.

    Returns:
        2-tuple: (the difficulty-stratified coverage violation, the number of samples, the empirical coverage and size of each difficulty).
    """
    if prediction_sets.shape[0] != len(labels):
        raise ValueError("Number of prediction sets must match number of labels")
    if not isinstance(strata_diff, list):
        raise TypeError("strata_diff must be a list")

    labels = labels.cpu()
    logits = logits.cpu()
    prediction_sets = prediction_sets.cpu()

    covered = prediction_sets[torch.arange(len(labels)), labels]
    set_sizes = prediction_sets.sum(dim=1)

    sorted_indices = torch.argsort(logits, dim=1, descending=True)
    topk = torch.zeros(len(labels), dtype=torch.int32)
    for i, (indices, label) in enumerate(zip(sorted_indices, labels)):
        topk[i] = (indices == label).nonzero(as_tuple=True)[0].item() + 1

    ccss_diff = {}
    diff_violation = -1

    for stratum in strata_diff:
        stratum_mask = (topk >= stratum[0]) & (topk <= stratum[1])
        stratum_size = stratum_mask.sum().item()

        ccss_diff[str(stratum)] = {'cnt': stratum_size}

        if stratum_size == 0:
            ccss_diff[str(stratum)].update({'cvg': 0, 'sz': 0})
            continue

        cvg = covered[stratum_mask].float().mean().item()
        sz = set_sizes[stratum_mask].float().mean().item()

        ccss_diff[str(stratum)].update({
            'cvg': round(cvg, 3),
            'sz': round(sz, 3)
        })
        stratum_violation = abs(1 - alpha - cvg)
        diff_violation = max(diff_violation, stratum_violation)

    return diff_violation, ccss_diff


@METRICS_REGISTRY_CLASSIFICATION.register()
def SSCV(prediction_sets, labels, alpha, stratified_size=[[0, 1], [2, 3], [4, 10], [11, 100], [101, 1000]]):
    """
    Size-stratified coverage violation (SSCV).
    
    Paper: Uncertainty Sets for Image Classifiers using Conformal Prediction (Angelopoulos et al., 2020)
    
    Link : https://iclr.cc/virtual/2021/spotlight/3435
    
    Args:
        prediction_sets (torch.Tensor): Boolean tensor of prediction sets (N x C).
        labels (torch.Tensor): Ground-truth labels (N,).
        alpha (float): User-guided confidence level (between 0 and 1).
        stratified_size (list): Coarse partitioning of possible set sizes.
            Each element should be a list [min_size, max_size] where:
            - min_size and max_size are non-negative integers
            - min_size <= max_size
            - Ranges should not overlap
    
    Returns:
        Int: The value of SSCV.
    
    """
    if len(prediction_sets) != len(labels):
        raise ValueError("The number of prediction sets must be equal to the number of labels.")

    if not isinstance(stratified_size, list) or not stratified_size:
        raise ValueError("stratified_size must be a non-empty list")

    labels = labels.cpu()
    prediction_sets = prediction_sets.cpu()

    size_array = prediction_sets.sum(dim=1)  # shape: (N,)
    correct_array = prediction_sets[torch.arange(len(labels)), labels]  # shape: (N,)

    sscv = -1
    for stratum in stratified_size:
        stratum_mask = (size_array >= stratum[0]) & (size_array <= stratum[1])
        stratum_size = stratum_mask.sum()

        if stratum_size > 0:
            stratum_coverage = correct_array[stratum_mask].float().mean()
            stratum_violation = torch.abs(1 - alpha - stratum_coverage).item()
            sscv = max(sscv, stratum_violation)
    return sscv


@METRICS_REGISTRY_CLASSIFICATION.register()
def WSC(features, prediction_sets, labels, delta=0.1, M=1000, test_fraction=0.75, random_state=2020, verbose=False):
    """
    Worst-Slice Coverage (WSC).
    
     Classification with Valid and Adaptive Coverage (Romano et al., 2020)
     Paper: Classification with Valid and Adaptive Coverage
     Link : https://proceedings.neurips.cc/paper/2020/hash/244edd7e85dc81602b7615cd705545f5-Abstract.html
     Code: https://github.com/msesia/arc/tree/d80d27519f18b11e7feaf8cf0da8827151af9ce3

    
    Args:
        features (torch.Tensor): Input features (N x D).
        prediction_sets (torch.Tensor): Boolean tensor of prediction sets (N x C).
        y (torch.Tensor): Ground-truth labels (N,).
        delta (float): Confidence level (between 0 and 1).
        M (int): Number of random projections.
        test_size (float): Proportion of test split.
        random_state (int): Random seed.
        verbose (bool): Whether to print progress.
    
     Returns:
         Float: the value of unbiased WSV.
    
    """

    if not 0 < delta < 1:
        raise ValueError("delta must be between 0 and 1")
    if not 0 < test_fraction < 1:
        raise ValueError("test_size must be between 0 and 1")
    if M <= 0:
        raise ValueError("M must be positive")

    if len(features.shape) != 2:
        raise ValueError(f"features must be 2D tensor, got shape {features.shape}")
    if len(prediction_sets.shape) != 2:
        raise ValueError(f"prediction_sets must be 2D tensor, got shape {prediction_sets.shape}")
    if len(labels.shape) != 1:
        raise ValueError(f"labels must be 1D tensor, got shape {labels.shape}")

    if features.shape[0] != len(labels):
        raise ValueError(
            f"Number of samples mismatch: features has {features.shape[0]} samples but labels has {len(labels)} samples")
    if features.shape[0] != prediction_sets.shape[0]:
        raise ValueError(
            f"Number of samples mismatch: features has {features.shape[0]} samples but prediction_sets has {prediction_sets.shape[0]} samples")
    if prediction_sets.shape[1] != len(torch.unique(labels)):
        raise ValueError(
            f"Number of classes mismatch: prediction_sets has {prediction_sets.shape[1]} classes but labels has {len(torch.unique(labels))} unique classes")

    features = features.cpu().numpy()
    labels_np = labels.cpu().numpy()
    covered = prediction_sets[np.arange(len(labels_np)), labels_np].cpu().numpy()

    X_train, X_test, y_train, y_test, covered_train, covered_test = train_test_split(features, labels_np, covered,
                                                                                     test_size=test_fraction,
                                                                                     random_state=random_state)
    # Find adversarial parameters
    wsc_star, v_star, a_star, b_star = calWSC(X_train, covered_train, y_train, delta=delta, M=M,
                                              random_state=random_state, verbose=verbose)
    # Estimate coverage
    coverage = wsc_vab(X_test, y_test, covered_test, v_star, a_star, b_star)
    return coverage.item()


def wsc_vab(featreus, labels, covered, v, a, b):
    z = np.dot(featreus, v)
    idx = (z >= a) & (z <= b)
    return np.mean(covered[idx])


def calWSC(X, y, covered, delta=0.1, M=1000, random_state=2020, verbose=True):
    rng = np.random.default_rng(random_state)
    n = len(y)

    def wsc_v(X, covered, delta, v):
        z = np.dot(X, v)
        z_order = np.argsort(z)
        z_sorted = z[z_order]
        cover_ordered = covered[z_order]

        ai_max = int(np.round((1.0 - delta) * n))
        ai_best = 0
        bi_best = n - 1
        cover_min = 1
        for ai in np.arange(0, ai_max):
            bi_min = np.minimum(ai + int(np.round(delta * n)), n - 1)
            coverage = np.cumsum(cover_ordered[ai:n]) / np.arange(1, n - ai + 1)
            coverage[np.arange(0, bi_min - ai)] = 1
            bi_star = ai + np.argmin(coverage)
            cover_star = coverage[bi_star - ai]
            if cover_star < cover_min:
                ai_best = ai
                bi_best = bi_star
                cover_min = cover_star
        return cover_min, z_sorted[ai_best], z_sorted[bi_best]

    def sample_sphere(n: int, p: int) -> np.ndarray:
        v = rng.normal(size=(p, n))
        v /= np.linalg.norm(v, axis=0)
        return v.T

    V = sample_sphere(M, p=X.shape[1])
    results = np.zeros((M, 4))

    iterator = tqdm(range(M)) if verbose else range(M)

    for m in iterator:
        wsc, a, b = wsc_v(X, covered, delta, V[m])
        results[m] = [wsc, a, b, m]

    idx_best = np.argmin(results[:, 0])
    wsc_star = results[idx_best, 0]
    a_star = results[idx_best, 1]
    b_star = results[idx_best, 2]
    v_star = V[int(results[idx_best, 3])]

    return wsc_star, v_star, a_star, b_star


@METRICS_REGISTRY_CLASSIFICATION.register()
def singleton_hit_ratio(prediction_sets, labels):
    if len(prediction_sets) == 0:
        raise AssertionError("The number of prediction set must be greater than 0.")
    n = len(prediction_sets)
    singletons = torch.sum(prediction_sets, dim=1) == 1
    covered = prediction_sets[torch.arange(len(labels)), labels]

    return torch.sum(singletons & covered).item() / n


def compute_p_values(cal_scores, test_scores, smooth=False):
    """
    Compute p-values for conformal prediction.

    Args:
        cal_scores (Tensor): Nonconformity scores from the calibration set, shape (n_cal,).
        test_scores (Tensor): Nonconformity scores for test samples across k classes, shape (n_test, k).
        smooth (bool): Whether to apply randomized smoothing when calibration scores equal test scores.

    Returns:
        Tensor: p-values for each test sample and class, shape (n_test, k)
    """

    n_cal = cal_scores.size(0)
    n_test, k = test_scores.size()

    cal_scores_expanded = cal_scores.view(1, n_cal, 1)
    test_scores_expanded = test_scores.view(n_test, 1, k)

    greater = (cal_scores_expanded > test_scores_expanded).sum(dim=1)
    equal = (cal_scores_expanded == test_scores_expanded).sum(dim=1)

    tau = torch.rand_like(equal, dtype=torch.float)

    if smooth:
        p_values = (greater + tau * (equal + 1)) / (n_cal + 1)
    else:
        p_values = (greater + (equal + 1)) / (n_cal + 1)
    return p_values


@METRICS_REGISTRY_CLASSIFICATION.register()
def pvalue_criterion_S(cal_scores, test_scores, smooth=False):
    """
    Paper: Criteria of efficiency for conformal prediction (Vovk et al., 2016)

    Sum criterion: measures efficiency by the average sum of the p-values.
                Smaller values are preferable.

    Args:
        cal_scores (Tensor): Nonconformity scores from the calibration set, shape (n_cal,).
        test_scores (Tensor): Nonconformity scores for test samples across k classes, shape (n_test, k).
        smooth (bool): Whether to apply randomized smoothing when calibration scores equal test scores.

    Returns:
        float: The average sum of the p-values across all test samples.
    """
    p_values = compute_p_values(cal_scores, test_scores, smooth)
    return p_values.sum(dim=1).mean()


@METRICS_REGISTRY_CLASSIFICATION.register()
def pvalue_criterion_N(cal_scores, test_scores, alpha, smooth=False):
    """
    Paper: Criteria of efficiency for conformal prediction (Vovk et al., 2016)

    Number criterion: uses the average size of the prediction sets.
                    Smaller values are preferable.

    Args:
        cal_scores (Tensor): Nonconformity scores from the calibration set, shape (n_cal,).
        test_scores (Tensor): Nonconformity scores for test samples across k classes, shape (n_test, k).
        alpha (float): The significance level.
        smooth (bool): Whether to apply randomized smoothing when calibration scores equal test scores.

    Returns:
        float: The average size of the prediction sets.
    """
    p_values = compute_p_values(cal_scores, test_scores, smooth)
    return (p_values > alpha).sum(dim=1).float().mean()


@METRICS_REGISTRY_CLASSIFICATION.register()
def pvalue_criterion_U(cal_scores, test_scores, smooth=False):
    """
    Paper: Criteria of efficiency for conformal prediction (Vovk et al., 2016)

    Unconfidence criterion: uses the average unconfidence over the test sequence, 
                            where the unconfidence for a test object x_i is the second largest p-value.
                            Smaller values are preferable.

    Args:
        cal_scores (Tensor): Nonconformity scores from the calibration set, shape (n_cal,).
        test_scores (Tensor): Nonconformity scores for test samples across k classes, shape (n_test, k).
        smooth (bool): Whether to apply randomized smoothing when calibration scores equal test scores.

    Returns:
        float: Mean of second-largest p-values across test samples.
    """
    p_values = compute_p_values(cal_scores, test_scores, smooth)
    n_test = p_values.size(0)

    max_classes = torch.argmax(p_values, dim=1)

    mask = torch.ones_like(p_values, dtype=bool)
    mask[torch.arange(n_test), max_classes] = False

    second_values = torch.where(mask, p_values, -torch.inf).amax(dim=1)
    return second_values.mean()


@METRICS_REGISTRY_CLASSIFICATION.register()
def pvalue_criterion_F(cal_scores, test_scores, smooth=False):
    """
    Paper: Criteria of efficiency for conformal prediction (Vovk et al., 2016)

    Fuzziness criterion: uses the average fuzziness where the fuzziness for a test object x_i is defined 
                        as the sum of all p_values apart from a largest one.
                        Smaller values are preferable.

    Args:
        cal_scores (Tensor): Nonconformity scores from the calibration set, shape (n_cal,).
        test_scores (Tensor): Nonconformity scores for test samples across k classes, shape (n_test, k).
        smooth (bool): Whether to apply randomized smoothing when calibration scores equal test scores.

    Returns:
        float: Mean sum of p-values minus the max p-value per sample.
    """
    p_values = compute_p_values(cal_scores, test_scores, smooth)
    return (p_values.sum(dim=1) - p_values.max(dim=1).values).mean()


@METRICS_REGISTRY_CLASSIFICATION.register()
def pvalue_criterion_M(cal_scores, test_scores, alpha, smooth=False):
    """
    Paper: Criteria of efficiency for conformal prediction (Vovk et al., 2016)

    Multiple criterion: uses the percentage of objects x_i in the test sequence 
                        for which the prediction set at significance level is multiple.
                        Smaller values are preferable.

    Args:
        cal_scores (Tensor): Nonconformity scores from the calibration set, shape (n_cal,).
        test_scores (Tensor): Nonconformity scores for test samples across k classes, shape (n_test, k).
        alpha (float): The significance level.
        smooth (bool): Whether to apply randomized smoothing when calibration scores equal test scores.

    Returns:
        float: Proportion of test samples with prediction set size > 1.
    """
    p_values = compute_p_values(cal_scores, test_scores, smooth)
    sizes = (p_values > alpha).sum(dim=1)
    return (sizes > 1).float().mean()


@METRICS_REGISTRY_CLASSIFICATION.register()
def pvalue_criterion_E(cal_scores, test_scores, alpha, smooth=False):
    """
    Paper: Criteria of efficiency for conformal prediction (Vovk et al., 2016)

    Excess criterion: uses the average amount the size of the prediction set exceeds 1.
                    Larger values are preferable.

    Args:
        cal_scores (Tensor): Nonconformity scores from the calibration set, shape (n_cal,).
        test_scores (Tensor): Nonconformity scores for test samples across k classes, shape (n_test, k).
        alpha (float): The significance level.
        smooth (bool): Whether to apply randomized smoothing when calibration scores equal test scores.

    Returns:
        float: Mean of (set size - 1), clamped at 0, across test samples.
    """
    p_values = compute_p_values(cal_scores, test_scores, smooth)
    sizes = (p_values > alpha).sum(dim=1)
    return torch.clamp(sizes - 1, min=0).float().mean()


@METRICS_REGISTRY_CLASSIFICATION.register()
def pvalue_criterion_OU(cal_scores, test_scores, test_labels, smooth=False):
    """
    Paper: Criteria of efficiency for conformal prediction (Vovk et al., 2016)

    Observed Unconfidence criterion: uses the average observed unconfidence over the test sequence, 
            where the observed unconfidence for a test example (x_i, y_i) is the largest p-value for the false labels.
            Smaller values are preferable.

    Args:
        cal_scores (Tensor): Nonconformity scores from the calibration set, shape (n_cal,).
        test_scores (Tensor): Nonconformity scores for test samples across k classes, shape (n_test, k).
        test_labels (Tensor): Ground-Truth labels for test samples.
        smooth (bool): Whether to apply randomized smoothing when calibration scores equal test scores.

    Returns:
        float: Mean of the highest p-value among incorrect classes per sample.
    """
    p_values = compute_p_values(cal_scores, test_scores, smooth)
    n_test = p_values.size(0)

    mask = torch.ones_like(p_values, dtype=bool)
    mask[torch.arange(n_test), test_labels] = False

    largest_false_values = torch.where(mask, p_values, -torch.inf).amax(dim=1)
    return largest_false_values.mean()


@METRICS_REGISTRY_CLASSIFICATION.register()
def pvalue_criterion_OF(cal_scores, test_scores, test_labels, smooth=False):
    """
    Paper: Criteria of efficiency for conformal prediction (Vovk et al., 2016)

    Observed Fuzziness criterion: uses the average sum of the pvalues for the false labels.
                                Smaller values are preferable.

    Args:
        cal_scores (Tensor): Nonconformity scores from the calibration set, shape (n_cal,).
        test_scores (Tensor): Nonconformity scores for test samples across k classes, shape (n_test, k).
        smooth (bool): Whether to apply randomized smoothing when calibration scores equal test scores.

    Returns:
        float: Mean sum of p-values excluding the true label per test sample.
    """
    p_values = compute_p_values(cal_scores, test_scores, smooth)
    n_test = p_values.size(0)

    mask = torch.ones_like(p_values, dtype=bool)
    mask[torch.arange(n_test), test_labels] = False

    sum_wrong_values = (p_values * mask).sum(dim=1)
    return sum_wrong_values.mean()


@METRICS_REGISTRY_CLASSIFICATION.register()
def pvalue_criterion_OM(cal_scores, test_scores, test_labels, alpha, smooth=False):
    """
    Paper: Criteria of efficiency for conformal prediction (Vovk et al., 2016)

    Observed Multiple criterion: uses the percentage of observed multiple predictions in the test sequence, 
            where an observed multiple prediction is defined to be a prediction set including a false label.
            Smaller values are preferable.

    Args:
        cal_scores (Tensor): Nonconformity scores from the calibration set, shape (n_cal,).
        test_scores (Tensor): Nonconformity scores for test samples across k classes, shape (n_test, k).
        test_labels (Tensor): Ground-Truth labels for test samples.
        alpha (float): The significance level.
        smooth (bool): Whether to apply randomized smoothing when calibration scores equal test scores.

    Returns:
        float: Proportion of test samples where prediction set contains at least one wrong class.
    """
    p_values = compute_p_values(cal_scores, test_scores, smooth)
    pred_sets = p_values > alpha

    pred_sets[torch.arange(pred_sets.size(0)), test_labels] = False
    return (pred_sets.sum(dim=1) > 0).float().mean()


@METRICS_REGISTRY_CLASSIFICATION.register()
def pvalue_criterion_OE(cal_scores, test_scores, test_labels, alpha, smooth=False):
    """
    Paper: Criteria of efficiency for conformal prediction (Vovk et al., 2016)

    Observed Excess criterion: uses the average number of false labels included 
                            in the prediction sets at significance level.
                            Smaller values are preferable.

    Args:
        cal_scores (Tensor): Nonconformity scores from the calibration set, shape (n_cal,).
        test_scores (Tensor): Nonconformity scores for test samples across k classes, shape (n_test, k).
        test_labels (Tensor): Ground-Truth labels for test samples.
        alpha (float): The significance level.
        smooth (bool): Whether to apply randomized smoothing when calibration scores equal test scores.

    Returns:
        float: Mean number of wrong classes in prediction sets across test samples.
    """
    p_values = compute_p_values(cal_scores, test_scores, smooth)
    pred_sets = p_values > alpha

    pred_sets[torch.arange(pred_sets.size(0)), test_labels] = False
    return pred_sets.sum(dim=1).float().mean()


class Metrics:

    def __call__(self, metric) -> Any:
        if metric not in METRICS_REGISTRY_CLASSIFICATION.registered_names():
            raise NameError(f"The metric: {metric} is not defined in TorchCP.")
        return METRICS_REGISTRY_CLASSIFICATION.get(metric)



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import math
import warnings

import pytest
import torch
from torch.utils.data import Dataset

from torchcp.classification.predictor import ClassConditionalPredictor
from torchcp.classification.score import LAC


@pytest.fixture
def mock_dataset():
    class MyDataset(Dataset):
        def __init__(self):
            self.x = torch.randn(100, 3)
            self.labels = torch.randint(0, 3, (100,))

        def __len__(self):
            return len(self.x)

        def __getitem__(self, idx):
            return self.x[idx], self.labels[idx]

    return MyDataset()


@pytest.fixture
def mock_model():
    class MockModel(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.param = torch.nn.Parameter(torch.tensor(1.0))

        def forward(self, x):
            return x

    return MockModel()


@pytest.fixture
def mock_score_function():
    return LAC(score_type="softmax")


@pytest.fixture
def predictor(mock_score_function, mock_model):
    return ClassConditionalPredictor(mock_score_function, mock_model)


def test_valid_initialization(predictor, mock_score_function, mock_model):
    assert predictor.score_function is mock_score_function
    assert predictor._model is mock_model
    assert not predictor._model.training
    assert predictor._device == next(mock_model.parameters()).device
    assert predictor._logits_transformation.temperature == 1.0
    assert predictor.q_hat is None


@pytest.mark.parametrize("temperature", [-0.1, -1.0])
def test_invalid_initialization(mock_score_function, mock_model, temperature):
    with pytest.raises(ValueError, match="temperature must be greater than 0."):
        ClassConditionalPredictor(mock_score_function, mock_model, temperature)


@pytest.mark.parametrize("alpha", [0.05])
def test_calculate_threshold(predictor, mock_score_function, alpha):
    logits = torch.randn(100, 3)
    zeros = torch.zeros(30, dtype=torch.long)
    ones = torch.ones(30, dtype=torch.long)
    twos = torch.full((40,), 2, dtype=torch.long)
    labels = torch.cat([zeros, ones, twos])

    predictor.calculate_threshold(logits, labels, None)

    predictor.calculate_threshold(logits, labels, alpha)

    num_classes = 3
    scores = mock_score_function(logits, labels)
    excepted_qhat = torch.zeros(num_classes)
    for label in range(num_classes):
        temp_scores = scores[labels == label]
        excepted_qhat[label] = torch.sort(temp_scores).values[math.ceil((temp_scores.shape[0] + 1) * (1 - alpha)) - 1]
    assert torch.equal(predictor.q_hat, excepted_qhat)

    logits = torch.randn(100, 3)
    labels = torch.cat([torch.zeros((90,)), torch.ones((10,))]).long()
    with warnings.catch_warnings(record=True) as warns:
        warnings.simplefilter("always")
        predictor.calculate_threshold(logits, labels, alpha)
    assert len(warns) == 2
    assert "The number of scores is 0" in str(warns[1].message)
    assert "The value of quantile exceeds 1" in str(warns[0].message)

    scores = mock_score_function(logits, labels)
    excepted_qhat = torch.zeros(num_classes)
    marginal_q_hat = torch.sort(scores).values[math.ceil((scores.shape[0] + 1) * (1 - alpha)) - 1]
    for label in range(num_classes):
        temp_scores = scores[labels == label]
        if len(temp_scores) == 0:
            excepted_qhat[label] = torch.inf
            continue
        quantile_value = math.ceil(temp_scores.shape[0] + 1) * (1 - alpha) / temp_scores.shape[0]
        if quantile_value > 1:
            excepted_qhat[label] = torch.inf
            continue
        excepted_qhat[label] = torch.sort(temp_scores).values[math.ceil((temp_scores.shape[0] + 1) * (1 - alpha)) - 1]
    assert torch.equal(predictor.q_hat, excepted_qhat)



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from math import sqrt

import pytest
import torch

from torchcp.graph.utils import compute_adj_knn


def test_invalid_compute_adj_knn():
    features = torch.tensor([
        [1.0, 0.0, 1.0, 0.0, 1.0],
        [0.0, 1.0, 0.0, 1.0, 0.0],
        [0.0, 0.0, 1.0, 0.0, 1.0],
        [0.0, 1.0, 0.0, 0.0, 1.0],
        [0.0, 1.0, 1.0, 0.0, 1.0]
    ])
    with pytest.raises(ValueError, match="The number of nodes cannot be less than k"):
        compute_adj_knn(features, 6)


def test_compute_adj_knn():
    features = torch.tensor([
        [1.0, 0.0, 1.0, 0.0, 1.0],
        [0.0, 1.0, 0.0, 1.0, 0.0],
        [0.0, 0.0, 1.0, 0.0, 1.0],
        [0.0, 1.0, 0.0, 0.0, 1.0],
        [0.0, 1.0, 1.0, 0.0, 1.0]
    ])
    k = 2
    knn_edge, knn_weights = compute_adj_knn(features, k)
    excepted_edge = torch.tensor([[0, 0, 1, 1, 2, 2, 3, 3, 4, 4],
                                  [2, 4, 3, 4, 0, 4, 2, 4, 2, 3]])
    excepted_weight = torch.tensor([2 / sqrt(6), 2 / 3, 1 / 2, 1 / sqrt(6), 2 / sqrt(6), 2 / sqrt(6),
                                    1 / 2, 2 / sqrt(6), 2 / sqrt(6), 2 / sqrt(6)])
    assert torch.equal(knn_edge, excepted_edge)
    assert torch.allclose(knn_weights, excepted_weight)



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import copy
from tqdm import tqdm

import torch
import torch.nn.functional as F

from torchcp.utils.common import get_device
from torchcp.classification.score import LAC, APS
from torchcp.classification.loss import ConfTrLoss
from torchcp.classification.predictor import SplitPredictor
from torchcp.graph.trainer.model import CFGNNModel


class CFGNNTrainer:
    """
    Method: Conformalized GNN
    Paper: Uncertainty Quantification over Graph with Conformalized Graph Neural Networks (Huang et al., 2023).
    Link: https://openreview.net/pdf?id=ygjQCOyNfh
    Github: https://github.com/snap-stanford/conformalized-gnn

    A class for training and evaluating a Conformalized GNN (CF-GNN) for graph classification tasks.
    The model uses a Graph Neural Network (GNN) as the backbone and integrates conformal prediction methods 
    for uncertainty quantification and model calibration.

    Args:
        model (torch.nn.Module): backbone model.
        graph_data (from torch_geometric.data import Data): 
            x (tensor): features of nodes.
            edge_index (Tensor): The edge index, shape (2, num_edges).
            edge_weight (Tensor, optional): The edge weights, shape (num_edges,).
            train_idx: The indices of the training nodes.
            val_idx: The indices of the validation nodes.
            calib_train_idx: The indices of the training nodes for CF-GNN.
        hidden_channels (int): Number of hidden channels for the CF-GNN layers.
        alpha (float, optional): The significance level for conformal prediction. Default is 0.1.
        optimizer_class (torch.optim.Optimizer): Optimizer class for temperature parameter
                Default: torch.optim.Adam
        optimizer_params (dict): Parameters passed to optimizer constructor
                Default: {'weight_decay': 5e-4, 'lr': 0.001}
    """

    def __init__(
            self,
            model,
            graph_data,
            hidden_channels=64,
            num_layers=2,
            alpha=0.1,
            optimizer_class: torch.optim.Optimizer = torch.optim.Adam,
            optimizer_params: dict = {'weight_decay': 5e-4, 'lr': 0.001},
            device=None):

        if graph_data is None:
            raise ValueError("graph_data cannot be None.")
        if model is None:
            raise ValueError("model cannot be None.")
        
        if device is not None:
            self._device = torch.device(device)
        else:
            self._device = get_device(model)

        self.graph_data = graph_data.to(self._device)

        self.num_classes = graph_data.y.max().item() + 1
        self.model = CFGNNModel(model, self.num_classes, hidden_channels, num_layers).to(self._device)

        self.optimizer = optimizer_class(
            self.model.parameters(),
            **optimizer_params
        )

        self.loss_fns = [F.cross_entropy,
                         ConfTrLoss(predictor=SplitPredictor(score_function=LAC(score_type="softmax")),
                                alpha=alpha,
                                fraction=0.5,
                                loss_type="classification",
                                target_size=0)]
        self.loss_weights = [1.0, 1.0]

        self.predictor = SplitPredictor(APS(score_type="softmax"))
        self.alpha = alpha

    def _train_each_epoch(self, epoch):
        """
        Trains the model for one epoch using the given data.

        Args:
            epoch: The current epoch number.
        """

        self.model.train()
        self.optimizer.zero_grad()

        logits = self.model(self.graph_data.x, self.graph_data.edge_index)
        loss = self.loss_fns[0](logits[self.graph_data.train_idx], self.graph_data.y[self.graph_data.train_idx])

        if epoch >= 1000:
            eff_loss = self.loss_fns[1](logits[self.graph_data.calib_train_idx],
                                        self.graph_data.y[self.graph_data.calib_train_idx])
            loss = self.loss_weights[0] * loss + self.loss_weights[1] * eff_loss

        loss.backward()
        self.optimizer.step()


    def _evaluate(self):
        """
        Evaluates the model's performance on the validation set.

        Returns:
            size (float): The average size of validation size.
        """
        self.model.eval()
        with torch.no_grad():
            logits = self.model(self.graph_data.x, self.graph_data.edge_index)

        val_perms = torch.randperm(self.graph_data.val_idx.size(0))
        valid_calib_idx = self.graph_data.val_idx[val_perms[:int(len(self.graph_data.val_idx) / 2)]]
        valid_test_idx = self.graph_data.val_idx[val_perms[int(len(self.graph_data.val_idx) / 2):]]

        self.predictor.calculate_threshold(logits[valid_calib_idx], self.graph_data.y[valid_calib_idx], self.alpha)
        pred_sets = self.predictor.predict_with_logits(logits[valid_test_idx])
        size = self.predictor._metric('average_size')(pred_sets, self.graph_data.y[valid_test_idx])

        return size

    def train(self, n_epochs=5000):
        """
        Trains the CF-GNN model for a specified number of epochs and returns the corrected logits.

        Args:
            n_epochs: The number of training epochs.

        Returns:
            model: The best model of CF-GNN.
        """

        best_valid_size = self.num_classes
        best_model_state = None

        for epoch in tqdm(range(n_epochs)):
            self._train_each_epoch(epoch)

            eff_valid = self._evaluate()

            if eff_valid < best_valid_size:
                best_valid_size = eff_valid
                best_model_state = copy.deepcopy(self.model.state_dict())

        if best_model_state is not None:
            self.model.load_state_dict(best_model_state)
        return self.model



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import torch


def compute_adj_knn(features, k=20):
    """
    Compute the k-nearest neighbor (k-NN) graph adjacency matrix for a given feature matrix.

    Args:
        features (torch.Tensor): A tensor of shape (N, D) where N is the number of nodes 
                                 and D is the dimensionality of the features.
        k (int): The number of nearest neighbors to consider for each node. Default is 20.

    Returns:
        tuple:
            - knn_edge (torch.Tensor): A tensor of shape (2, E) containing the edge indices 
                                      of the k-NN graph in COO format (source, target).
            - knn_weights (torch.Tensor): A tensor of shape (E,) containing the similarity 
                                          weights of the edges in the k-NN graph.
    """
    if features.shape[0] < k:
        raise ValueError(
            "The number of nodes cannot be less than k.")
    features_normalized = features / features.norm(dim=1, keepdim=True)
    sims = torch.mm(features_normalized, features_normalized.t())
    sims[(torch.arange(len(sims)), torch.arange(len(sims)))] = 0

    topk_values, topk_indices = torch.topk(sims, k, dim=1)

    adj_knn = torch.zeros_like(sims).to(features.device)
    rows = torch.arange(sims.shape[0]).unsqueeze(1).to(features.device)
    adj_knn[rows, topk_indices] = topk_values

    knn_edge = torch.nonzero(adj_knn).T
    knn_weights = adj_knn[knn_edge[0, :], knn_edge[1, :]]

    return knn_edge, knn_weights



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

__all__ = ["ConfTr"]

import torch
import torch.nn.functional as F

from torchcp.classification.loss.confts import ConfTSLoss


class ConfTrLoss(ConfTSLoss):
    """
    Conformal Training (ConfTr) Loss Implementation.
    
    A method for training neural networks with built-in conformal prediction guarantees.
    Optimizes models to output efficient prediction sets while maintaining coverage.

    Args:
        predictor (torchcp.classification.Predictor): An instance of the CP predictor class.
        alpha (float): The significance level for each training batch.
        fraction (float): The fraction of the calibration set in each training batch.
            Must be a value in (0, 1).
        soft_qunatile (bool, optional): Whether to use soft quantile. Default is True.
        epsilon (float, optional): A temperature value. Default is 1e-4.
        loss_type (str): The selected (multi-selected) loss functions, which can be
            "valid", "classification", "probs", "coverage".
        target_size (int, optional): Optional: 0 | 1. Default is 1.
        loss_transform (str, optional): A transform for loss. Default is "square".
            Can be "square", "abs", or "log".
        
    Examples::
        >>> predictor = torchcp.classification.SplitPredictor()
        >>> conftr = ConfTr(predictor=predictor, alpha=0.05, fraction=0.2, loss_type="valid")
        >>> logits = torch.randn(100, 10)
        >>> labels = torch.randint(0, 2, (100,))
        >>> loss = conftr(logits, labels)
        >>> loss.backward()
        
    Reference:
        Stutz et al. "Learning Optimal Conformal Classifiers", ICLR 2021, https://arxiv.org/abs/2110.09192
        
        Github: https://github.com/google-deepmind/conformal_training
    """

    def __init__(self, predictor, alpha, fraction, soft_qunatile=True, epsilon=1e-4, loss_type="valid",
                 target_size=1, loss_transform="square"):

        super(ConfTrLoss, self).__init__(predictor, alpha, fraction, soft_qunatile)

        if loss_type not in ["valid", "classification", "probs", "coverage", "cfgnn"]:
            raise ValueError(
                'loss_type should be a value in ["valid", "classification", "probs", "coverage"].')
        if target_size not in [0, 1]:
            raise ValueError("target_size should be 0 or 1.")
        if loss_transform not in ["square", "abs", "log"]:
            raise ValueError('loss_transform should be a value in ["square", "abs", "log"].')
        if epsilon <= 0:
            raise ValueError("epsilon must be greater than 0.")

        self.predictor = predictor
        self.alpha = alpha
        self.fraction = fraction
        self.loss_type = loss_type
        self.target_size = target_size
        self.soft_qunatile = soft_qunatile
        self.epsilon = epsilon

        if loss_transform == "square":
            self.transform = torch.square
        elif loss_transform == "abs":
            self.transform = torch.abs
        elif loss_transform == "log":
            self.transform = torch.log
        self.loss_functions_dict = {"valid": self.__compute_hinge_size_loss,
                                    "probs": self.__compute_probabilistic_size_loss,
                                    "coverage": self.__compute_coverage_loss,
                                    "classification": self.__compute_classification_loss,
                                    }

    def compute_loss(self, test_scores, test_labels, tau):
        pred_sets = torch.sigmoid((tau - test_scores) / self.epsilon)
        return self.loss_functions_dict[self.loss_type](pred_sets, test_labels)

    def __compute_hinge_size_loss(self, pred_sets, labels):
        return torch.mean(
            self.transform(
                torch.relu(torch.sum(pred_sets, dim=1) - self.target_size)))

    def __compute_probabilistic_size_loss(self, pred_sets, labels):
        classes = pred_sets.shape[1]
        one_hot_labels = torch.unsqueeze(torch.eye(classes).to(pred_sets.device), dim=0)
        repeated_confidence_sets = pred_sets.unsqueeze(2).repeat(1, 1, classes)
        loss = one_hot_labels * repeated_confidence_sets + \
               (1 - one_hot_labels) * (1 - repeated_confidence_sets)
        loss = torch.prod(loss, dim=1)
        return torch.sum(loss, dim=1)

    def __compute_coverage_loss(self, pred_sets, labels):
        one_hot_labels = F.one_hot(labels, num_classes=pred_sets.shape[1])

        # Compute the mean of the sum of confidence_sets multiplied by one_hot_labels
        loss = torch.mean(torch.sum(pred_sets * one_hot_labels, dim=1)) - (1 - self.alpha)

        # Apply the transform function (you need to define this)
        transformed_loss = self.transform(loss)

        return transformed_loss

    def __compute_classification_loss(self, pred_sets, labels):
        # Convert labels to one-hot encoding
        one_hot_labels = F.one_hot(labels, num_classes=pred_sets.shape[1]).float()
        loss_matrix = torch.eye(pred_sets.shape[1], device=pred_sets.device)
        # Calculate l1 and l2 losses
        l1 = (1 - pred_sets) * one_hot_labels * loss_matrix[labels]
        l2 = pred_sets * (1 - one_hot_labels) * loss_matrix[labels]

        # Calculate the total loss
        loss = torch.sum(torch.maximum(l1 + l2, torch.zeros_like(l1, device=pred_sets.device)), dim=1)

        # Return the mean loss
        return torch.mean(loss)



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from .abs import ABS
from .norabs import NorABS, DifficultyEstimator
from .cqr import CQR
from .cqrfm import CQRFM
from .cqrm import CQRM
from .cqrr import CQRR
from .r2ccp import R2CCP
from .sign import Sign


# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import torch

from torchcp.classification.score.aps import APS


class RAPS(APS):
    """
    Method: Regularized Adaptive Prediction Sets 
    Paper: Uncertainty Sets for Image Classifiers using Conformal Prediction (Angelopoulos et al., 2020)
    Link: https://arxiv.org/abs/2009.14193
    
    Args:
        penalty (float): The weight of regularization. When penalty = 0, RAPS=APS.
        kreg (int, optional): The rank of regularization which is an integer in [0, classes_num]. Default is 0.
        score_type (str, optional): The type of score to use. Default is "softmax".
        randomized (bool, optional): Whether to use randomized scores. Default is True.
        
    Examples::
        >>> raps = RAPS(penalty=0.1, kreg=1, score_type="softmax", randomized=True)
        >>> probs = torch.tensor([[0.1, 0.4, 0.5], [0.3, 0.3, 0.4]])
        >>> scores_all = raps._calculate_all_label(probs)
        >>> print(scores_all)
        >>> scores_single = raps._calculate_single_label(probs, torch.tensor([2, 1]))
        >>> print(scores_single)
        
    """

    def __init__(self, score_type="softmax", randomized=True, penalty=0, kreg=0):

        super().__init__(score_type=score_type, randomized=randomized)
        if penalty < 0:
            raise ValueError("The parameter 'penalty' must be a nonnegative value.")

        if type(kreg) != int or kreg < 0:
            raise TypeError("The parameter 'kreg' must be a nonnegative integer.")
        self.__penalty = penalty
        self.__kreg = kreg

    def _calculate_all_label(self, probs):
        indices, ordered, cumsum = self._sort_sum(probs)
        if self.randomized:
            U = torch.rand(probs.shape, device=probs.device)
        else:
            U = torch.zeros_like(probs)
        reg = torch.maximum(self.__penalty * (torch.arange(1, probs.shape[-1] + 1, device=probs.device) - self.__kreg),
                            torch.tensor(0, device=probs.device))
        ordered_scores = cumsum - ordered * U + reg
        _, sorted_indices = torch.sort(indices, descending=False, dim=-1)
        scores = ordered_scores.gather(dim=-1, index=sorted_indices)
        return scores

    def _calculate_single_label(self, probs, label):

        indices, ordered, cumsum = self._sort_sum(probs)
        if self.randomized:
            U = torch.rand(indices.shape[0], device=probs.device)
        else:
            U = torch.zeros(indices.shape[0], device=probs.device)
        idx = torch.where(indices == label.view(-1, 1))
        reg = torch.maximum(self.__penalty * (idx[1] + 1 - self.__kreg), torch.tensor(0).to(probs.device))
        scores = cumsum[idx] - U * ordered[idx] + reg
        return scores



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import torch
import torch.nn as nn
from torch import Tensor


class TemperatureScalingModel(nn.Module):
    """
    Model wrapper for temperature scaling in conformal prediction.
    
    Applies learnable temperature parameter to logits to improve
    calibration for conformal prediction.
    
    Args:
        base_model (nn.Module): Pre-trained model to be calibrated
        temperature (float): Initial temperature value (default: 1.0)
    
    Shape:
        - Input: Same as base_model input
        - Output: (batch_size, num_classes) scaled logits
    
    Examples:
        >>> base_model = resnet18(pretrained=True)
        >>> model = TemperatureScalingModel(base_model)
        >>> inputs = torch.randn(10, 3, 224, 224)
        >>> scaled_logits = model(inputs)
        
    Reference:
        Guo et al. "On Calibration of Modern Neural Networks", ICML 2017, https://arxiv.org/abs/1706.04599
    """

    def __init__(self, base_model: nn.Module, temperature: float = 1.0):
        super().__init__()
        if temperature <= 0:
            raise ValueError("Temperature must be positive")
        self.base_model = base_model
        self.temperature = nn.Parameter(torch.ones(1) * temperature)

        # Freeze base model parameters
        self.freeze_base_model()

    def freeze_base_model(self):
        """Freeze all parameters in base model"""
        for param in self.base_model.parameters():
            param.requires_grad = False
        self.base_model.eval()

    def is_base_model_frozen(self) -> bool:
        """Check if base model parameters are frozen"""
        return not any(p.requires_grad for p in self.base_model.parameters())

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass with temperature scaling.
        
        Args:
            x: Input tensor
            
        Returns:
            Temperature scaled logits
        """
        with torch.no_grad():  # Ensure no gradients flow through base model
            logits = self.base_model(x)

        return logits / self.temperature

    def get_temperature(self) -> float:
        """Get current temperature value."""
        return self.temperature.item()

    def set_temperature(self, temp: float) -> None:
        """Set temperature value."""
        if temp <= 0:
            raise ValueError("Temperature must be positive")
        with torch.no_grad():
            self.temperature.fill_(temp)

    def train(self, mode: bool = True):
        """
        Override train method to ensure base_model stays in eval mode
        
        Args:
            mode: boolean to specify training mode
        """
        super().train(mode)  # Set training mode for TemperatureScalingModel
        self.base_model.eval()  # Keep base_model in eval mode
        return self


class OrdinalClassifier(nn.Module):
    """
    Ordinal Classifier 

    Args:
        classifier (nn.Module): A PyTorch classifier model.
        phi (str, optional): The transformation function for the classifier output. Default is "abs". Options are "abs" and "square".
        varphi (str, optional): The transformation function for the cumulative sum. Default is "abs". Options are "abs" and "square".

    Attributes:
        classifier (nn.Module): The classifier model.
        phi (str): The transformation function for the classifier output.
        varphi (str): The transformation function for the cumulative sum.
        phi_function (callable): The transformation function applied to the classifier output.
        varphi_function (callable): The transformation function applied to the cumulative sum.

    Methods:
        forward(x):
            Forward pass of the model.
    
    Examples::
        >>> classifier = nn.Linear(10, 5)
        >>> model = OrdinalClassifier(classifier, phi="abs", varphi="abs")
        >>> x = torch.randn(3, 10)
        >>> output = model(x)
        >>> print(output)
        
    Reference:
        DEY et al., 2023. "Conformal Prediction Sets for Ordinal Classification", https://proceedings.neurips.cc/paper_files/paper/2023/hash/029f699912bf3db747fe110948cc6169-Abstract-Conference.html
    """

    def __init__(self, classifier, phi="abs", varphi="abs"):
        super().__init__()
        self.classifier = classifier
        self.phi = phi
        self.varphi = varphi

        phi_options = {"abs": torch.abs, "square": torch.square}
        varphi_options = {"abs": lambda x: -torch.abs(x), "square": lambda x: -torch.square(x)}

        if phi not in phi_options:
            raise NotImplementedError(f"phi function '{self.phi}' is not implemented. Options are 'abs' and 'square'.")
        if varphi not in varphi_options:
            raise NotImplementedError(
                f"varphi function '{self.varphi}' is not implemented. Options are 'abs' and 'square'.")

        self.phi_function = phi_options[phi]
        self.varphi_function = varphi_options[phi]

    def forward(self, x):
        """
        Forward pass of the model.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The output tensor after applying the classifier, phi_function, and varphi_function.
        """
        if x.shape[1] <= 2:
            raise ValueError("The input dimension must be greater than 2.")

        x = self.classifier(x)

        # a cumulative summation
        x = torch.cat((x[:, :1], self.phi_function(x[:, 1:])), dim=1)

        x = torch.cumsum(x, dim=1)

        # the unimodal distribution
        x = self.varphi_function(x)
        return x


class SurrogateCPModel(nn.Module):
    """
    This model wraps a given base model and adds a linear layer on top of its final feature output.
    The base model's parameters are frozen to prevent updates during training, so only the added
    linear layer is trainable.

    Args:
        base_model (nn.Module): Pre-trained model
    
    Shape:
        - Input: Same as base_model input
        - Output: (batch_size, num_classes) 1 - logits
    
    Examples:
        >>> base_model = resnet18(pretrained=True)
        >>> model = SurrogateCPModel(base_model)
        >>> inputs = torch.randn(10, 3, 224, 224)
        >>> logits = model(inputs)
    """

    def __init__(self, base_model: nn.Module):
        super().__init__()
        self.base_model = base_model
        self.linear = nn.Linear(in_features=base_model.fc.out_features, 
                                out_features=base_model.fc.out_features,
                                bias=False)

        # Freeze base model parameters
        self.freeze_base_model()

    def freeze_base_model(self):
        """Freeze all parameters in base model"""
        for param in self.base_model.parameters():
            param.requires_grad = False
        self.base_model.eval()

    def is_base_model_frozen(self) -> bool:
        """Check if base model parameters are frozen"""
        return not any(p.requires_grad for p in self.base_model.parameters())

    def forward(self, x: Tensor) -> Tensor:
        with torch.no_grad():  # Ensure no gradients flow through base model
            logits = self.base_model(x)

        return 1 - self.linear(logits)

    def train(self, mode: bool = True):
        super().train(mode)  # Set training mode for TemperatureScalingModel
        self.base_model.eval()  # Keep base_model in eval mode
        return self


# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from typing import Any

import torch

from torchcp.utils.registry import Registry

METRICS_REGISTRY_REGRESSION = Registry("METRICS")


@METRICS_REGISTRY_REGRESSION.register()
def coverage_rate(prediction_intervals, y_truth):
    """
    Calculates the coverage rate of prediction intervals.
    
    Args:
        prediction_intervals (torch.Tensor): A tensor of shape (batch_size, num_intervals * 2),
                                             where each interval has a lower and upper bound.
        y_truth (torch.Tensor): A tensor of ground truth values of shape (batch_size,).
    
    Returns:
        torch.Tensor: The coverage rate, representing the proportion of ground truth values
                      within the specified prediction intervals.
    
    """
    num_columns = prediction_intervals.shape[-1]
    if num_columns % 2 != 0:
        raise ValueError(f"The number of columns in prediction_intervals must be even, but got {num_columns}")

    if len(prediction_intervals.shape) == 2:
        prediction_intervals = prediction_intervals.unsqueeze(1)
    if len(y_truth.shape) == 1:
        y_truth = y_truth.unsqueeze(1)

    condition = torch.zeros_like(y_truth, dtype=torch.bool)

    for i in range(num_columns // 2):
        lower_bound = prediction_intervals[..., 2 * i]
        upper_bound = prediction_intervals[..., 2 * i + 1]
        condition |= torch.bitwise_and(y_truth >= lower_bound, y_truth <= upper_bound)

    coverage_rate = torch.sum(condition, dim=0).cpu() / y_truth.shape[0]
    return coverage_rate.item()


@METRICS_REGISTRY_REGRESSION.register()
def average_size(prediction_intervals):
    """
    Computes the average size of prediction intervals.
    
    Args:
        prediction_intervals (torch.Tensor): A tensor of shape (batch_size, num_intervals * 2),
                                             where each interval has a lower and upper bound.
    
    Returns:
        torch.Tensor: The average size of the prediction intervals across all samples.
    """
    num_columns = prediction_intervals.shape[-1]
    if num_columns % 2 != 0:
        raise ValueError(f"The number of columns in prediction_intervals must be even, but got {num_columns}")

    size = torch.abs(prediction_intervals[..., 1::2] - prediction_intervals[..., 0::2]).sum(dim=-1)
    average_size = size.mean(dim=0).cpu().item()

    return average_size


class Metrics:
    def __call__(self, metric) -> Any:
        if metric not in METRICS_REGISTRY_REGRESSION.registered_names():
            raise NameError(f"The metric: {metric} is not defined in TorchCP.")
        return METRICS_REGISTRY_REGRESSION.get(metric)



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import torch
from torchsort import soft_rank, soft_sort

REG_STRENGTH = 0.1  # Regularization strength used in soft sorting for smoothness
B = 50  # A parameter controlling the smoothness of the soft indicator function


class UniformMatchingLoss(torch.nn.Module):
    """
    A custom loss function that calculates the discrepancy in the sorting of input tensor x.
    It measures how far off each element is from its ideal position in a sorted sequence.
    """

    def __init__(self):
        super().__init__()

    def forward(self, x):
        """
        Forward pass function that computes the loss value for the given input tensor x.
        
        Args:
            x: A tensor (usually the model's output) with shape (batch_size,).
                This represents the predicted scores or values for each sample.
        
        Returns:
            out: A scalar loss value representing the inconsistency in element sorting.
        """
        batch_size = len(x)
        if batch_size == 0:
            return 0
        x_sorted = soft_sort(x.unsqueeze(dim=0), regularization_strength=REG_STRENGTH)
        i_seq = torch.arange(1.0, 1.0 + batch_size, device=x.device) / batch_size
        out = torch.max(torch.abs(i_seq - x_sorted))
        return out


class UncertaintyAwareLoss(torch.nn.Module):
    """
    A loss function used for conformalized uncertainty-aware training of deep multi-class classifiers

    Examples:
        >>> conflearn_loss_fn = ConfLearnLoss()
        >>> output = torch.randn(100, 10)
        >>> target = torch.randint(0, 2, (100,))
        >>> Z_batch = torch.randint(0, 2, (100,))
        >>> loss = conflearn_loss_fn(output, target, Z_batch)
        >>> loss.backward()

    Reference:
        Einbinder et al. "Training Uncertainty-Aware Classifiers with Conformalized Deep Learning" (2022), https://arxiv.org/abs/2205.05878
    """

    def __init__(self):
        super(UncertaintyAwareLoss, self).__init__()

        self.layer_prob = torch.nn.Softmax(dim=1)
        self.criterion_scores = UniformMatchingLoss()

    def forward(self, output, target):
        """
        Forward pass of the conformal loss function. The loss is computed by iterating over different groupings in Z_batch,
        applying the conformal loss for each group, and averaging the loss over all groups.

        Args:
            output (torch.Tensor): The model's output logits (predictions before softmax).
            target (torch.Tensor): The ground truth labels.

        Returns:
            torch.Tensor: The computed loss for the given batch.
        """
        return self.compute_loss(output, target)

    def compute_loss(self, y_train_pred, y_train_batch):
        """
        Computes the conformal loss for a given batch of predictions and ground truth.

        Args:
            y_train_pred (torch.Tensor): The model's predicted logits for the batch.
            y_train_batch (torch.Tensor): The ground truth labels for the batch.

        Returns:
            torch.Tensor: The conformal loss for the batch.
        """
        train_proba = self.layer_prob(y_train_pred)
        train_scores = self.__compute_scores_diff(
            train_proba, y_train_batch)
        train_loss_scores = self.criterion_scores(train_scores)
        return train_loss_scores

    def __compute_scores_diff(self, proba_values, Y_values):
        """
        Computes the non-conformity scores based on the predicted probabilities and the true labels.
        This score measures how different the predicted probabilities are from the actual labels.

        Args:
            proba_values (torch.Tensor): The predicted probabilities for the batch (after softmax).
            Y_values (torch.Tensor): The ground truth labels for the batch.

        Returns:
            torch.Tensor: The computed non-conformity scores for each sample in the batch.
        """
        device = proba_values.device
        n, K = proba_values.shape
        proba_values = proba_values + 1e-6 * \
                       torch.rand(proba_values.shape, dtype=float, device=device)
        proba_values = proba_values / torch.sum(proba_values, 1)[:, None]
        ranks_array_t = soft_rank(-proba_values,
                                  regularization_strength=REG_STRENGTH) - 1
        prob_sort_t = -soft_sort(-proba_values,
                                 regularization_strength=REG_STRENGTH)
        Z_t = prob_sort_t.cumsum(dim=1)

        ranks_t = torch.gather(
            ranks_array_t, 1, Y_values.reshape(n, 1)).flatten()
        prob_cum_t = self.__soft_indexing(Z_t, ranks_t)
        prob_final_t = self.__soft_indexing(prob_sort_t, ranks_t)
        scores_t = 1.0 - prob_cum_t + prob_final_t * \
                   torch.rand(n, dtype=float, device=device)

        return scores_t

    def __soft_indicator(self, x, a, b=B):
        """
        Soft indicator function, which is a smoothed version of a step function.
        This is used for soft indexing in the loss computation to smooth out the discrete jumps.

        Args:
            x (torch.Tensor): The tensor of indices to compute the indicator function over.
            a (torch.Tensor): The rank tensor, indicating the position.
            b (float): Regularization strength, controlling the smoothness of the indicator function.

        Returns:
            torch.Tensor: The soft indicator values.
        """
        out = torch.sigmoid(b * (x - a + 0.5)) - (torch.sigmoid(b * (x - a - 0.5)))
        out = out / (torch.sigmoid(torch.tensor(b * 0.5)) - torch.sigmoid(-torch.tensor(b * 0.5)))
        return out

    def __soft_indexing(self, z, rank):
        """
        Soft indexing operation used to calculate weighted sums based on predicted probabilities
        and the ranks of true labels. This smooths the selection of indices during loss computation.

        Args:
            z (torch.Tensor): The cumulative sorted probabilities.
            rank (torch.Tensor): The ranks corresponding to the true labels.

        Returns:
            torch.Tensor: The weighted sum of indexed values, representing the soft loss.
        """
        n = len(rank)
        K = z.shape[1]
        I = torch.tile(torch.arange(K, device=z.device), (n, 1))
        weight = self.__soft_indicator(I.T, rank).T
        weight = weight * z
        return weight.sum(dim=1)



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import pytest
import torch
import torch.nn as nn
from torch.utils.data import TensorDataset, DataLoader
import tempfile
import os

import pytest
import torch
import torch.nn as nn
from torch.utils.data import DataLoader, TensorDataset
import os
import tempfile

from torchcp.classification.trainer.base_trainer import BaseTrainer, Trainer

# Mock model for testing
class MockModel(nn.Module):
    def __init__(self):
        super().__init__()
        self.linear = nn.Linear(10, 2)
        
    def forward(self, x):
        return self.linear(x)

# Concrete implementation of BaseTrainer for testing abstract class
class ConcreteTrainer(BaseTrainer):
    def train(self, train_loader, val_loader=None, **kwargs):
        return self.model

# Fixtures
@pytest.fixture
def mock_model():
    return MockModel()

@pytest.fixture
def mock_data():
    # Create mock dataset
    x = torch.randn(100, 10)  # 100 samples, 10 features
    y = torch.randint(0, 2, (100,))  # Binary classification
    dataset = TensorDataset(x, y)
    return DataLoader(dataset, batch_size=16)

@pytest.fixture
def base_trainer(mock_model):
    return ConcreteTrainer(mock_model)

@pytest.fixture
def trainer(mock_model):
    return Trainer(mock_model)

# Tests for BaseTrainer
class TestBaseTrainer:
    def test_train_abstract_method(self):
        class FailTrainer(BaseTrainer):
            pass  
        
        with pytest.raises(TypeError, match="Can't instantiate abstract class"):
            FailTrainer(MockModel())
            
    def test_train_implementation(self, base_trainer, mock_data):
        model = base_trainer.train(mock_data)
        assert isinstance(model, torch.nn.Module)
        
        model = base_trainer.train(mock_data, mock_data)
        assert isinstance(model, torch.nn.Module)
        
        model = base_trainer.train(mock_data, extra_param=True)
        assert isinstance(model, torch.nn.Module)
        
        
    def test_initialization(self, mock_model):
        # Test with default device
        trainer = ConcreteTrainer(mock_model)
        assert trainer.device == torch.device('cpu')
        
        # Test with specified device
        cpu_trainer = ConcreteTrainer(mock_model, device=torch.device('cpu'))
        assert cpu_trainer.device == torch.device('cpu')
        
        # Test verbose setting
        assert trainer.verbose == True
        quiet_trainer = ConcreteTrainer(mock_model, verbose=False)
        assert quiet_trainer.verbose == False

    def test_invalid_initialization(self):
        with pytest.raises(ValueError, match="Model cannot be None"):
            ConcreteTrainer(None)

    def test_model_save_load(self, base_trainer):
        with tempfile.NamedTemporaryFile(suffix='.pt', delete=False) as tmp:
            try:
                # Save model
                base_trainer.save_model(tmp.name)
                assert os.path.exists(tmp.name)
                
                # Change model parameters
                original_params = {name: param.clone() for name, param in base_trainer.model.named_parameters()}
                for param in base_trainer.model.parameters():
                    nn.init.constant_(param, 0.0)
                
                # Load model and verify parameters are restored
                base_trainer.load_model(tmp.name)
                for name, param in base_trainer.model.named_parameters():
                    assert torch.allclose(param, original_params[name])
            finally:
                os.unlink(tmp.name)


    def test_train_pass_statement(self):
        """直接测试基类中train方法的pass语句"""
        class TempTrainer(BaseTrainer):
            def train(self, train_loader, val_loader=None, **kwargs):
                return super().train(train_loader, val_loader, **kwargs)
        
        model = MockModel()
        trainer = TempTrainer(model)
        data_loader = DataLoader(TensorDataset(torch.randn(10, 10), torch.zeros(10)))
        
        result = BaseTrainer.train.__get__(trainer)(data_loader)
        
        assert result is None
    
    
# Tests for Trainer
class TestTrainer:
    def test_initialization(self, trainer):
        assert isinstance(trainer.optimizer, torch.optim.Adam)
        assert isinstance(trainer.loss_fn, nn.CrossEntropyLoss)

    def test_calculate_loss(self, trainer):
        output = torch.randn(4, 2)  # 4 samples, 2 classes
        target = torch.tensor([0, 1, 0, 1])
        loss = trainer.calculate_loss(output, target)
        assert isinstance(loss, torch.Tensor)
        assert loss.ndim == 0  # Scalar tensor

    def test_train_epoch(self, trainer, mock_data):
        loss = trainer.train_epoch(mock_data)
        assert isinstance(loss, float)
        assert loss > 0  # Loss should be positive

    def test_validate(self, trainer, mock_data):
        loss = trainer.validate(mock_data)
        assert isinstance(loss, float)
        assert loss > 0

    def test_full_training(self, trainer, mock_data):
        # Test training with validation
        trained_model = trainer.train(
            train_loader=mock_data,
            val_loader=mock_data,
            num_epochs=2
        )
        assert isinstance(trained_model, nn.Module)
        
        # Test training without validation
        trained_model = trainer.train(
            train_loader=mock_data,
            num_epochs=2
        )
        assert isinstance(trained_model, nn.Module)

    def test_model_save_load(self, trainer):
        with tempfile.NamedTemporaryFile(suffix='.pt', delete=False) as tmp:
            try:
                # Get initial parameters
                initial_params = {name: param.clone() for name, param in trainer.model.named_parameters()}
                
                # Save model
                trainer.save_model(tmp.name)
                assert os.path.exists(tmp.name)
                
                # Change parameters
                for param in trainer.model.parameters():
                    nn.init.constant_(param, 0.0)
                
                # Load and verify
                trainer.load_model(tmp.name)
                for name, param in trainer.model.named_parameters():
                    assert torch.allclose(param, initial_params[name])
            finally:
                os.unlink(tmp.name)

    def test_training_with_early_stopping(self, trainer, mock_data):
        """Test that the best model is saved during training with validation"""
        initial_state = {name: param.clone() for name, param in trainer.model.named_parameters()}
        
        # Train with validation to trigger best model saving
        trainer.train(
            train_loader=mock_data,
            val_loader=mock_data,
            num_epochs=3
        )
        
        # Verify model parameters have changed
        current_state = {name: param.clone() for name, param in trainer.model.named_parameters()}
        any_param_changed = False
        for name in initial_state:
            if not torch.allclose(initial_state[name], current_state[name]):
                any_param_changed = True
                break
        assert any_param_changed, "Model parameters should have changed during training"
        





# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import torch
import torch.optim as optim

from torchcp.regression.loss import QuantileLoss
from torchcp.regression.score.base import BaseScore
from torchcp.regression.utils import build_regression_model


class CQR(BaseScore):
    """
    Conformalized Quantile Regression (CQR) 
    This score function allows for calculating scores and generating prediction intervals
    using quantile regression model.

    Reference:
        Paper: Conformalized Quantile Regression (Romano et al., 2019)
        Link: https://proceedings.neurips.cc/paper_files/paper/2019/file/5103c3584b063c431bd1268e9b5e76fb-Paper.pdf
        Github: https://github.com/yromano/cqr
    """

    def __init__(self):
        super().__init__()

    def __call__(self, predicts, y_truth):
        """
        Calculates the non-conformity scores for predictions.

        Non-conformity scores are defined as the absolute deviation between the true values 
        and the predicted values. These scores are used to evaluate the quality of prediction 
        intervals in conformal prediction.

        Args:
            predicts (torch.Tensor): Tensor of predicted quantile intervals, shape (batch_size, 2).
            y_truth (torch.Tensor): Tensor of true target values, shape (batch_size,).

        Returns:
            torch.Tensor: Tensor of non-conformity scores, shape (batch_size,).
        """
        if len(predicts.shape) == 2:
            predicts = predicts.unsqueeze(1)
        if len(y_truth.shape) == 1:
            y_truth = y_truth.unsqueeze(1)
        return torch.maximum(predicts[..., 0] - y_truth, y_truth - predicts[..., 1])

    def generate_intervals(self, predicts_batch, q_hat):
        """
        Generates prediction intervals for a batch of predictions.

        Prediction intervals are constructed by adding and subtracting the calibrated
        thresholds (:attr:`q_hat`) from the predicted values.

        Args:
            predicts_batch (torch.Tensor): A batch of predictions, shape (batch_size, ...).
            q_hat (torch.Tensor): Calibrated thresholds, shape (num_thresholds,).

        Returns:
            torch.Tensor: Prediction intervals, shape (batch_size, num_thresholds, 2).
                          The last dimension contains the lower and upper bounds of the intervals.
        """
        if len(predicts_batch.shape) == 2:
            predicts_batch = predicts_batch.unsqueeze(1)
        prediction_intervals = predicts_batch.new_zeros((predicts_batch.shape[0], q_hat.shape[0], 2))
        prediction_intervals[..., 0] = predicts_batch[..., 0] - q_hat.view(1, q_hat.shape[0], 1)
        prediction_intervals[..., 1] = predicts_batch[..., 1] + q_hat.view(1, q_hat.shape[0], 1)
        return prediction_intervals

    def train(self, train_dataloader, **kwargs):
        """
        Trains the model on provided training data with :math:`[alpha/2, 1-alpha/2]` quantile regression loss.

        Args:
            train_dataloader (torch.utils.data.DataLoader): DataLoader providing training data.
            **kwargs: Additional training parameters.
                - model (torch.nn.Module, optional): Model to be trained; defaults to the model passed to the predictor.
                - criterion (callable, optional): Loss function for training. If not provided, uses :func:`QuantileLoss`.
                - alpha (float, optional): Significance level (e.g., 0.1) for quantiles, required if :attr:`criterion` is None.
                - epochs (int, optional): Number of training epochs. Default is :math:`100`.
                - lr (float, optional): Learning rate for optimizer. Default is :math:`0.01`.
                - optimizer (torch.optim.Optimizer, optional): Optimizer for training; defaults to :func:`torch.optim.Adam`.
                - verbose (bool, optional): If True, displays training progress. Default is True.

        Raises:
            ValueError: If :attr:`criterion` is not provided and :attr:`alpha` is not specified.
        """
        device = kwargs.get('device', None)
        model = kwargs.get('model',
                           build_regression_model("NonLinearNet")(next(iter(train_dataloader))[0].shape[1], 2, 64,
                                                                  0.5).to(device))
        criterion = kwargs.get('criterion', None)

        if criterion is None:
            alpha = kwargs.get('alpha', None)
            if alpha is None:
                raise ValueError("When 'criterion' is not provided, 'alpha' must be specified.")
            quantiles = [alpha / 2, 1 - alpha / 2]
            criterion = QuantileLoss(quantiles)

        epochs = kwargs.get('epochs', 100)
        lr = kwargs.get('lr', 0.01)
        optimizer = kwargs.get('optimizer', optim.Adam(model.parameters(), lr=lr))
        verbose = kwargs.get('verbose', True)

        self._basetrain(model, epochs, train_dataloader, criterion, optimizer, verbose)
        return model



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

"""Utilities for score scaling."""
import math

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import DataLoader, RandomSampler, TensorDataset
from tqdm import tqdm


class BaseScaler:
    """Straight through scaler."""

    def fit(self, X, y):
        pass

    def predict(self, X):
        return X


class LogisticRegression(nn.Module):
    def __init__(self, dtype=torch.float32):
        super(LogisticRegression, self).__init__()
        self.linear = nn.Linear(1, 1)
        self.dtype = dtype

        self.linear.weight.data = self.linear.weight.data.to(dtype)
        self.linear.bias.data = self.linear.bias.data.to(dtype)

    def forward(self, x):
        x = x.to(self.dtype)
        return torch.sigmoid(self.linear(x))


class PlattScaler(BaseScaler):
    """Parametric calibration using logistic regression on uncalibrated scores."""

    def __init__(self, **kwargs):
        super(PlattScaler, self).__init__()
        self.clf = LogisticRegression()
        self.optimizer = optim.SGD(self.clf.parameters(), lr=0.01)
        self.loss_fn = nn.BCELoss()
        self.num_epochs = kwargs.get('num_epochs', 100)

    def fit(self, X, y):
        X = X.reshape(-1, 1)
        y = y.reshape(-1, 1)

        for epoch in range(self.num_epochs):
            self.optimizer.zero_grad()
            outputs = self.clf(X)
            loss = self.loss_fn(outputs, y.to(self.clf.dtype))
            loss.backward()
            self.optimizer.step()

    def predict(self, X):
        self.clf.eval()
        with torch.no_grad():
            scaled = self.clf(X.reshape(-1, 1))
        return scaled.reshape(X.shape)


class BinningScaler(BaseScaler):
    """Non-parametric equal-mass histogram regression on uncalibrated scores."""

    def __init__(self, n_bins=20):
        if n_bins <= 0:
            raise ValueError("n_bins must be positive")
        self.n_bins = n_bins

    def fit(self, X, y):
        X = X.reshape(-1)
        y = y.reshape(-1)

        # Split scores into equal mass bins.
        quantiles = torch.linspace(0, 1, self.n_bins + 1).to(X.dtype)
        bins = torch.quantile(X, quantiles)

        bin_ids = torch.searchsorted(bins[1:-1], X)

        # Count empirical accuracy in each bin.
        bin_true = torch.bincount(bin_ids, weights=y, minlength=len(bins))[:-1]
        bin_count = torch.bincount(bin_ids, minlength=len(bins))[:-1]
        bin_prob = bin_true / bin_count

        # Store bins and bin probs.
        self.bins = bins
        self.bin_prob = bin_prob

    def predict(self, X):
        bin_ids = torch.searchsorted(self.bins[1:-1], X.reshape(-1))
        return self.bin_prob[bin_ids].reshape(X.shape)


class PlattBinningScaler(BaseScaler):
    """Combined parametric + non-parametric calibration (Kumar et. al., 2019)"""

    def __init__(self, *args, n_bins=20, **kwargs):
        self.platt = PlattScaler(*args, **kwargs)
        self.binning = BinningScaler(n_bins=n_bins)

    def fit(self, X, y):
        # Split calibration set in two for scaling and binning.
        N = len(X) // 2
        X_scale, y_scale = X[:N], y[:N]
        X_bin, y_bin = X[N:], y[N:]

        # Fit Platt scaler.
        self.platt.fit(X_scale, y_scale)

        # Fit binning scaler on Platt scaled scores.
        self.binning.fit(self.platt.predict(X_bin), y_bin)

    def predict(self, X):
        return self.binning.predict(self.platt.predict(X))


class RecurrentScaler(BaseScaler, nn.Module):
    """RNN calibration of C given sequential scores."""

    def __init__(
            self,
            hidden_size=64,
            num_layers=2,
            num_iters=100,
            batch_size=128,
            dropout=0.0,
            target='set',
            verbose=True
    ):
        super(RecurrentScaler, self).__init__()
        if target not in ['set', 'item']:
            raise ValueError(f"target must be one of ['set', 'item'], but got {target}")
        self.num_iters = num_iters
        self.batch_size = batch_size
        self.target = target
        self.verbose = verbose
        self.rnn = nn.RNN(
            input_size=1,
            hidden_size=hidden_size,
            num_layers=num_layers,
            dropout=dropout,
            batch_first=True)
        self.hidden_to_output = nn.Linear(hidden_size, 1)

    def forward(self, X):
        out = self.rnn(X.unsqueeze(-1))[0]
        out = self.hidden_to_output(out).squeeze(-1)
        return out

    def fit(self, X, y):
        X = X.to(torch.float32)
        # If self.target = 'set', then we compute the set-based labels.
        if self.target == 'set':
            y = 1 - torch.cumprod(1 - y, axis=-1)

        dataset = TensorDataset(X, y)
        sampler = RandomSampler(
            dataset, num_samples=self.num_iters * self.batch_size)
        loader = DataLoader(
            dataset, sampler=sampler, batch_size=self.batch_size)

        # Train loop.
        optimizer = optim.Adam(self.parameters(), lr=1e-3)
        self.train()
        if self.verbose:
            loader = tqdm(loader)
        for X_batch, y_batch in loader:
            out = self.forward(X_batch)
            loss = F.binary_cross_entropy_with_logits(out, y_batch)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

    def predict(self, X):
        X = X.to(torch.float32)
        self.eval()
        with torch.no_grad():
            out = torch.sigmoid(self.forward(X))
        return out



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import pytest
import torch
import torch.nn.functional as F

from torch_geometric.data import Data
from torch_geometric.nn import GCNConv

from torchcp.classification.score import LAC, APS
from torchcp.classification.loss import ConfTrLoss
from torchcp.classification.predictor import SplitPredictor
from torchcp.graph.trainer import CFGNNTrainer
from torchcp.graph.trainer.model import CFGNNModel


@pytest.fixture
def device():
    return torch.device('cuda' if torch.cuda.is_available() else 'cpu')


@pytest.fixture
def mock_graph_data(device):
    num_nodes = 200
    x = torch.randn(num_nodes, 10)

    edge_index = torch.tensor([
        [0, 1, 2, 3, 4],
        [1, 2, 3, 4, 0]
    ])
    y = torch.randint(0, 10, (num_nodes,))
    rand_perm = torch.randperm(num_nodes)
    train_idx = rand_perm[:20]
    val_idx = rand_perm[20:40]
    calib_train_idx = rand_perm[40:60]

    return Data(x=x, edge_index=edge_index, y=y,
                num_nodes=num_nodes,
                train_idx=train_idx,
                val_idx=val_idx,
                calib_train_idx=calib_train_idx).to(device)


@pytest.fixture
def mock_model(device):
    class MockModel(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.param = torch.nn.Parameter(torch.tensor(1.0))

        def forward(self, x, edge_index=None):
            return x

    return MockModel().to(device)


@pytest.fixture
def mock_cfgnn_model(mock_model, mock_graph_data):
    return CFGNNTrainer(mock_model, mock_graph_data)


def test_initialization(mock_model, mock_graph_data):
    cf_trainer = CFGNNTrainer(mock_model, mock_graph_data)

    assert cf_trainer.graph_data is mock_graph_data
    assert cf_trainer._device == mock_graph_data.x.device
    assert cf_trainer.num_classes == 10

    assert isinstance(cf_trainer.model, CFGNNModel)
    assert cf_trainer.model.base_model is mock_model
    assert cf_trainer.model.model.convs[0].in_channels == 10
    assert cf_trainer.model.model.convs[0].out_channels == 64
    assert len(cf_trainer.model.model.convs) == 2
    assert cf_trainer.model.model.convs[1].in_channels == 64
    assert cf_trainer.model.model.convs[1].out_channels == 10

    assert isinstance(cf_trainer.optimizer, torch.optim.Adam)
    assert cf_trainer.optimizer.param_groups[0]['weight_decay'] == 5e-4
    assert cf_trainer.optimizer.param_groups[0]['lr'] == 0.001

    assert isinstance(cf_trainer.loss_fns, list)
    assert cf_trainer.loss_fns[0] == F.cross_entropy
    assert isinstance(cf_trainer.loss_fns[1], ConfTrLoss)
    assert isinstance(cf_trainer.loss_fns[1].predictor, SplitPredictor)
    assert isinstance(cf_trainer.loss_fns[1].predictor.score_function, LAC)
    assert cf_trainer.loss_fns[1].alpha == 0.1
    assert cf_trainer.loss_fns[1].fraction == 0.5
    assert cf_trainer.loss_fns[1].loss_type == "classification"
    assert cf_trainer.loss_fns[1].target_size == 0

    assert cf_trainer.loss_weights[0] == 1.0
    assert cf_trainer.loss_weights[1] == 1.0
    assert isinstance(cf_trainer.predictor, SplitPredictor)
    assert isinstance(cf_trainer.predictor.score_function, APS)
    assert cf_trainer.predictor.score_function.score_type == "softmax"
    assert cf_trainer.alpha == 0.1

    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    cf_trainer = CFGNNTrainer(mock_model, mock_graph_data, device=device)
    assert cf_trainer._device == device


def test_invalid_initialization(mock_graph_data, mock_model):
    with pytest.raises(ValueError, match="graph_data cannot be None"):
        CFGNNTrainer(mock_model, None)

    with pytest.raises(ValueError, match="model cannot be None"):
        CFGNNTrainer(None, mock_graph_data)


def test_train_each_epoch( mock_cfgnn_model):
    mock_cfgnn_model._train_each_epoch(500)
    mock_cfgnn_model._train_each_epoch(2000)


def test_evaluate(mock_cfgnn_model, mock_graph_data, mock_model):
    torch.manual_seed(42)
    mock_cfgnn_model.model = mock_model
    size = mock_cfgnn_model._evaluate()

    torch.manual_seed(42)
    mock_model.eval()
    with torch.no_grad():
        logits = mock_model(mock_graph_data.x, mock_graph_data.edge_index)

    val_perms = torch.randperm(mock_graph_data.val_idx.size(0))
    valid_calib_idx = mock_graph_data.val_idx[val_perms[:int(len(mock_graph_data.val_idx) / 2)]]
    valid_test_idx = mock_graph_data.val_idx[val_perms[int(len(mock_graph_data.val_idx) / 2):]]

    mock_cfgnn_model.predictor.calculate_threshold(logits[valid_calib_idx], mock_graph_data.y[valid_calib_idx], 0.1)
    pred_sets = mock_cfgnn_model.predictor.predict_with_logits(logits[valid_test_idx])
    except_size = mock_cfgnn_model.predictor._metric('average_size')(pred_sets, mock_graph_data.y[valid_test_idx])
    assert size == except_size


def test_train(mock_cfgnn_model):
    model = mock_cfgnn_model.train(10)
    assert model is mock_cfgnn_model.model



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import torch
import torch.optim as optim

from torchcp.regression.loss import R2ccpLoss
from torchcp.regression.score.base import BaseScore
from torchcp.regression.utils import build_regression_model


class R2CCP(BaseScore):
    """
    Regression-to-Classification Conformal Prediction.
    
    This method converting regression to a classification problem and 
    then use CP for classification to obtain CP sets for regression.

    Args:
        midpoints (torch.Tensor): the midpoints of the equidistant bins.
        
    Reference:
        Paper: Conformal Prediction via Regression-as-Classification (Etash Guha et al., 2021)
        Link: https://neurips.cc/virtual/2023/80610
        Github: https://github.com/EtashGuha/R2CCP
    """

    def __init__(self, midpoints, device=None):
        super().__init__()
        if device is not None:
            self._device = torch.device(device)
        else:
            self._device = midpoints.device
        self.midpoints = midpoints.to(self._device)

    def train(self, train_dataloader, **kwargs):
        """
        Trains regression-to-classification model with the R2ccpLoss.

        Args:
            train_dataloader (torch.utils.data.DataLoader): DataLoader for training data.
            kwargs: Additional parameters for training.
                - model (torch.nn.Module, optional): Model to be trained; defaults to the model passed to the predictor.
                - epochs (int, optional): Number of training epochs. Default is :math:`100`.
                - p (float, optional): Probability parameter for the loss function. Default is :math:`0.5`.
                - tau (float, optional): Threshold parameter for the loss function. Default is :math:`0.2`.
                - criterion (callable, optional): Loss function; defaults to :class:`R2ccpLoss`.
                - lr (float, optional): Learning rate. Default is :math:`1e-4`.
                - weight_decay (float, optional): Weight decay for the optimizer. Default is :math:`1e-4`.
                - optimizer (torch.optim.Optimizer, optional): Optimizer; defaults to :func:`torch.optim.AdamW`.
                - verbose (bool, optional): If True, displays training progress. Default is True.
                
        """
        device = kwargs.get('device', self._device)
        self._device = device
        model = kwargs.get('model', build_regression_model("NonLinearNet")(next(iter(train_dataloader))[0].shape[1],
                                                                           len(self.midpoints), 1000, 0).to(
            self._device))
        epochs = kwargs.get('epochs', 100)
        p = kwargs.get('p', 0.5)
        tau = kwargs.get('tau', 0.2)
        criterion = kwargs.get('criterion', R2ccpLoss(p, tau, self.midpoints))
        lr = kwargs.get('lr', 1e-4)
        weight_decay = kwargs.get('weight_decay', 1e-4)
        optimizer = kwargs.get('optimizer', optim.AdamW(model.parameters(), lr=lr, weight_decay=weight_decay))
        verbose = kwargs.get('verbose', True)

        self._basetrain(model, epochs, train_dataloader, criterion, optimizer, verbose)
        return model

    def __call__(self, predicts, y_truth):
        interval = self.__find_interval(self.midpoints, y_truth)
        scores = self.__calculate_linear_interpolation(interval, predicts, y_truth, self.midpoints)
        return - scores  # since r2ccp calculates conformity score instead of nonconformity score

    def generate_intervals(self, predicts_batch, q_hat):
        q_hat = - q_hat  # since r2ccp calculates conformity score instead of nonconformity score
        K = predicts_batch.shape[1]
        N = predicts_batch.shape[0]
        midpoints_expanded = self.midpoints.unsqueeze(0).expand(N, K)
        left_points = midpoints_expanded[:, :-1]
        right_points = midpoints_expanded[:, 1:]
        left_predicts = predicts_batch[:, :-1]
        right_predicts = predicts_batch[:, 1:]
        q_hat_expanded = torch.ones((predicts_batch.shape[0], predicts_batch.shape[1] - 1),
                                    device=self._device) * q_hat

        mask1 = (left_predicts >= q_hat_expanded) & (right_predicts >= q_hat_expanded)
        mask2 = (left_predicts <= q_hat_expanded) & (right_predicts <= q_hat_expanded)
        mask3 = (left_predicts < q_hat_expanded) & (right_predicts > q_hat_expanded)
        mask4 = (left_predicts > q_hat_expanded) & (right_predicts < q_hat_expanded)

        prediction_intervals = torch.zeros((N, 2 * (K - 1)), device=self._device)
        prediction_intervals[:, 0::2][mask1] = left_points[mask1]
        prediction_intervals[:, 1::2][mask1] = right_points[mask1]
        prediction_intervals[:, 0::2][mask2] = 0
        prediction_intervals[:, 1::2][mask2] = 0

        delta_midpoints = right_points - left_points
        prediction_intervals[:, 0::2][mask3] = right_points[mask3] - delta_midpoints[mask3] * \
                                               (right_predicts[mask3] - q_hat_expanded[mask3]) / (
                                                       right_predicts[mask3] - left_predicts[mask3])
        prediction_intervals[:, 1::2][mask3] = right_points[mask3]

        prediction_intervals[:, 0::2][mask4] = left_points[mask4]
        prediction_intervals[:, 1::2][mask4] = left_points[mask4] - delta_midpoints[mask4] * \
                                               (left_predicts[mask4] - q_hat_expanded[mask4]) / (
                                                       right_predicts[mask4] - left_predicts[mask4])

        return prediction_intervals

    def __find_interval(self, midpoints, y_truth):
        """
        Finds the interval index for each ground truth value based on midpoints, 
        i.e., midpoints[interval[i]] <= y_truth[i] < midpoints[interval[i+1]]

        Args:
            midpoints (torch.Tensor): Midpoints of the equidistant bins.
            y_truth (torch.Tensor): Ground truth values, shape (batch_size,).

        Returns:
            torch.Tensor: Interval indices for each ground truth value, shape (batch_size,).
        """
        interval = torch.zeros_like(y_truth, dtype=torch.long).to(self._device)

        for i in range(len(midpoints) + 1):
            if i == 0:
                mask = y_truth < midpoints[i]
                interval[mask] = -1
            elif i < len(midpoints):
                mask = (y_truth >= midpoints[i - 1]) & (y_truth < midpoints[i])
                interval[mask] = i - 1
            else:
                mask = y_truth >= midpoints[-1]
                interval[mask] = len(midpoints)
        return interval

    def __calculate_linear_interpolation(self, interval, predicts, y_truth, midpoints):
        """
        Calculates scores through linear interpolation within intervals.

        Args:
            interval (torch.Tensor): Interval indices for each data point, shape (batch_size,).
            predicts (torch.Tensor): Predicted bin probabilities, shape (batch_size, num_bins).
            y_truth (torch.Tensor): Ground truth values, shape (batch_size,).
            midpoints (torch.Tensor): Midpoints of the equidistant bins.

        Returns:
            torch.Tensor: Scores for each data point, shape (batch_size,).
        """
        midpoints = midpoints.repeat((y_truth.shape[0], 1))
        left_points = midpoints[torch.arange(y_truth.shape[0]), (interval) % midpoints.shape[1]]
        right_points = midpoints[torch.arange(y_truth.shape[0]), (interval + 1) % midpoints.shape[1]]

        left_predicts = predicts[torch.arange(y_truth.shape[0]), (interval) % midpoints.shape[1]]
        right_predicts = predicts[torch.arange(y_truth.shape[0]), (interval + 1) % midpoints.shape[1]]
        scores = ((y_truth - left_points) * right_predicts + (right_points - y_truth) * left_predicts) / (
                right_points - left_points)
        scores = torch.where(interval == -1, predicts[:, 0], scores)
        scores = torch.where(interval == len(midpoints), predicts[:, -1], scores)

        if len(scores.shape) == 1:
            scores = scores.unsqueeze(-1)

        return scores



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from math import sqrt

import pytest
import torch
from torch_geometric.data import Data

from torchcp.classification.score import LAC
from torchcp.graph.score import SNAPS


@pytest.fixture
def graph_data():
    x = torch.tensor([
        [1.0, 2.0],
        [3.0, 4.0],
        [5.0, 6.0]
    ], dtype=torch.float32)

    edge_index = torch.tensor([
        [0, 1, 1, 2],
        [1, 0, 2, 1]
    ], dtype=torch.long)

    y = torch.tensor([0, 1, 0], dtype=torch.long)

    data = Data(x=x, edge_index=edge_index, edge_weight=None, y=y)

    return data


@pytest.fixture
def base_score_function():
    return LAC(score_type="softmax")


@pytest.mark.parametrize("xi", [-0.1, 1.1])
@pytest.mark.parametrize("mu", [-0.1, 1.1])
def test_invalid_lambda_muues(graph_data, base_score_function, xi, mu):
    with pytest.raises(ValueError, match="The parameter 'xi' must be a value between 0 and 1."):
        SNAPS(graph_data, base_score_function, xi=xi)

    with pytest.raises(ValueError, match="The parameter 'mu' must be a value between 0 and 1."):
        SNAPS(graph_data, base_score_function, mu=mu)

    with pytest.raises(ValueError, match="The summation of 'xi' and 'mu' must not be greater than 1."):
        SNAPS(graph_data, base_score_function, xi=0.6, mu=0.6)

    with pytest.raises(ValueError, match="knn_edge and features cannot both be non-None"):
        SNAPS(graph_data, base_score_function, knn_edge=graph_data.edge_index, features=graph_data.x)


def test_valid_initialization(graph_data, base_score_function):
    score_function = SNAPS(graph_data, base_score_function, xi=0.3, mu=0.3, k=2)
    assert score_function._xi == 0.3
    assert score_function._mu == 0.3

    score_function = SNAPS(graph_data, base_score_function, features=graph_data.x, k=2)
    excepted_adjknn = torch.tensor([[0, 11 / (5 * sqrt(5)), 17 / sqrt(305)],
                                    [11 / (5 * sqrt(5)), 0, 39 / (5 * sqrt(61))],
                                    [17 / sqrt(305), 39 / (5 * sqrt(61)), 0]])
    assert torch.allclose(score_function._adj_knn.to_dense(), excepted_adjknn)


def test_knn_processing(graph_data, base_score_function):
    def are_sparse_tensors_equal(tensor1, tensor2):
        if tensor1.shape != tensor2.shape:
            return False
        if tensor1.device != tensor2.device:
            return False
        if not torch.equal(tensor1.coalesce().indices(), tensor2.coalesce().indices()):
            return False
        if not torch.equal(tensor1.coalesce().values(), tensor2.coalesce().values()):
            return False
        return True

    knn_edge = torch.tensor([
        [0, 1, 2],
        [1, 2, 0]
    ], dtype=torch.long)
    knn_weight = torch.tensor([0.5, 0.5, 0.5], dtype=torch.float32)

    adj_knn = torch.sparse_coo_tensor(
        knn_edge,
        knn_weight,
        (graph_data.num_nodes, graph_data.num_nodes))
    knn_degs = torch.matmul(adj_knn, torch.ones((adj_knn.shape[0])))

    score_function = SNAPS(graph_data, base_score_function, knn_edge=knn_edge, knn_weight=knn_weight)
    assert are_sparse_tensors_equal(score_function._adj_knn, adj_knn)
    assert torch.equal(score_function._knn_degs, knn_degs)

    score_function = SNAPS(graph_data, base_score_function, knn_edge=None, knn_weight=knn_weight)
    excepted_adjknn = torch.tensor([[0, 11 / (5 * sqrt(5)), 17 / sqrt(305)],
                                    [11 / (5 * sqrt(5)), 0, 39 / (5 * sqrt(61))],
                                    [17 / sqrt(305), 39 / (5 * sqrt(61)), 0]])
    assert torch.allclose(score_function._adj_knn.to_dense(), excepted_adjknn)

    score_function = SNAPS(graph_data, base_score_function, knn_edge=knn_edge, knn_weight=None)

    knn_weight = torch.ones(knn_edge.shape[1])
    adj_knn = torch.sparse_coo_tensor(
        knn_edge,
        knn_weight,
        (graph_data.num_nodes, graph_data.num_nodes))
    knn_degs = torch.matmul(adj_knn, torch.ones((adj_knn.shape[0])))
    assert are_sparse_tensors_equal(score_function._adj_knn, adj_knn)
    assert torch.equal(score_function._knn_degs, knn_degs)


def test_snaps_call_without_labels(graph_data, base_score_function):
    knn_edge = torch.tensor([
        [0, 1, 2],
        [1, 2, 0]
    ], dtype=torch.long)
    knn_weight = torch.tensor([0.5, 0.5, 0.5], dtype=torch.float32)
    snaps = SNAPS(graph_data, base_score_function, xi=0.4, mu=0.25, knn_edge=knn_edge,
                  knn_weight=knn_weight)

    logits = torch.tensor([
        [1.0, 0.5],
        [0.2, 0.8],
        [0.4, 0.6]
    ], dtype=torch.float32)

    base_scores = base_score_function(logits)
    similarity_scores = torch.tensor([
        [base_scores[1, 0], base_scores[1, 1]],
        [base_scores[2, 0], base_scores[2, 1]],
        [base_scores[0, 0], base_scores[0, 1]]
    ], dtype=torch.float32)
    neigh_scores = torch.tensor([
        [base_scores[1, 0], base_scores[1, 1]],
        [(base_scores[0, 0] + base_scores[2, 0]) / 2, (base_scores[0, 1] + base_scores[2, 1]) / 2],
        [base_scores[1, 0], base_scores[1, 1]]
    ], dtype=torch.float32)

    expected_scores = 0.35 * base_scores + 0.4 * similarity_scores + 0.25 * neigh_scores

    scores = snaps(logits)
    assert torch.allclose(scores, expected_scores, atol=1e-5)


def test_snaps_call_with_labels(graph_data, base_score_function):
    knn_edge = torch.tensor([
        [0, 1, 2],
        [1, 2, 0]
    ], dtype=torch.long)
    knn_weight = torch.tensor([0.5, 0.5, 0.5], dtype=torch.float32)
    snaps = SNAPS(graph_data, base_score_function, xi=0.4, mu=0.25, knn_edge=knn_edge,
                  knn_weight=knn_weight)

    logits = torch.tensor([
        [1.0, 0.5],
        [0.2, 0.8],
        [0.4, 0.6]
    ], dtype=torch.float32)

    labels = torch.tensor([0, 1, 0], dtype=torch.long)

    scores = snaps(logits, labels)
    expected_scores = snaps(logits)
    assert torch.allclose(scores, expected_scores[torch.arange(3), labels], atol=1e-5)



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from .base_trainer import BaseTrainer
from .conftr_trainer import ConfTrTrainer
from .confts_trainer import ConfTSTrainer
from .model_zoo import TemperatureScalingModel
from .ts_trainer import TSTrainer
from .ua_trainer import UncertaintyAwareTrainer
from .ordinal_trainer import OrdinalTrainer
from .scpo_trainer import SCPOTrainer


# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import pytest
import torch
from torch.utils.data import DataLoader, TensorDataset


@pytest.fixture
def dummy_data():
    """
    Fixture to provide dummy data for testing.
    """
    x_train = torch.rand((100, 10))
    y_train = torch.rand((100, 1))
    train_dataset = TensorDataset(x_train, y_train)
    train_dataloader = DataLoader(train_dataset, batch_size=16)

    x_test = torch.rand((20, 10))
    y_test = torch.rand((20, 1))
    test_dataset = TensorDataset(x_test, y_test)
    test_dataloader = DataLoader(test_dataset, batch_size=16)

    return train_dataloader, test_dataloader



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import math

import pytest
import torch
from torch.utils.data import Dataset

from torchcp.classification.predictor import SplitPredictor
from torchcp.classification.predictor.base import BasePredictor
from torchcp.classification.score import LAC
from torchcp.classification.utils.metrics import Metrics


@pytest.fixture
def mock_dataset():
    class MyDataset(Dataset):
        def __init__(self):
            self.x = torch.randn(100, 3)
            self.labels = torch.randint(0, 3, (100,))

        def __len__(self):
            return len(self.x)

        def __getitem__(self, idx):
            return self.x[idx], self.labels[idx]

    return MyDataset()


@pytest.fixture
def mock_model():
    class MockModel(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.param = torch.nn.Parameter(torch.tensor(1.0))

        def forward(self, x):
            return x

    return MockModel()


@pytest.fixture
def mock_score_function():
    return LAC(score_type="softmax")


@pytest.fixture
def predictor(mock_score_function, mock_model):
    return SplitPredictor(mock_score_function, mock_model)


def test_base_predictor_abstractmethod(mock_score_function):
    class Predictor(BasePredictor):
        def __init__(self, score_function):
            super().__init__(score_function)

    predictor = Predictor(mock_score_function)
    with pytest.raises(NotImplementedError):
        predictor.calibrate(None, 0.1)
    with pytest.raises(NotImplementedError):
        predictor.predict(None)


def test_valid_initialization(predictor, mock_score_function, mock_model):
    assert predictor.score_function is mock_score_function
    assert predictor._model is mock_model
    assert not predictor._model.training
    assert predictor._device == next(mock_model.parameters()).device
    assert predictor._logits_transformation.temperature == 1.0


def test_initialization_device(mock_score_function, mock_model):
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    predictor = SplitPredictor(mock_score_function, mock_model, device=device)
    assert predictor._device == device
    

@pytest.mark.parametrize("temperature", [-0.1, -1.0])
def test_invalid_initialization(mock_score_function, mock_model, temperature):
    with pytest.raises(ValueError, match="temperature must be greater than 0."):
        SplitPredictor(mock_score_function, mock_model, temperature)


@pytest.mark.parametrize("alpha", [0.1, 0.05])
def test_calibrate(predictor, mock_dataset, mock_score_function, mock_model, alpha):
    cal_dataloader = torch.utils.data.DataLoader(mock_dataset, batch_size=40)
    predictor.calibrate(cal_dataloader, None)

    predictor.calibrate(cal_dataloader, alpha)

    mock_model.eval()
    logits = mock_model(mock_dataset.x)
    scores = mock_score_function(logits, mock_dataset.labels)
    excepted_qhat = torch.sort(scores).values[math.ceil((scores.shape[0] + 1) * (1 - alpha)) - 1]
    assert predictor.q_hat == excepted_qhat

def test_invalid_calibrate_model(mock_score_function, mock_dataset):
    predictor = SplitPredictor(mock_score_function, None)
    cal_dataloader = torch.utils.data.DataLoader(mock_dataset, batch_size=40)
    with pytest.raises(ValueError, match="Model is not defined"):
        predictor.calibrate(cal_dataloader, 0.1)


@pytest.mark.parametrize("alpha", [0.1, 0.05])
def test_calculate_threshold(predictor, mock_score_function, alpha):
    logits = torch.randn(100, 3)
    labels = torch.randint(0, 3, (100,))

    predictor.calculate_threshold(logits, labels, alpha)

    scores = mock_score_function(logits, labels)
    excepted_qhat = torch.sort(scores).values[math.ceil((scores.shape[0] + 1) * (1 - alpha)) - 1]
    assert predictor.q_hat == excepted_qhat

    predictor.calculate_threshold(logits, labels, None)


@pytest.mark.parametrize("q_hat", [0.5, 0.7])
def test_predict(predictor, mock_score_function, mock_model, mock_dataset, q_hat):
    predictor.q_hat = q_hat
    pred_sets = predictor.predict(mock_dataset.x)

    logits = mock_model(mock_dataset.x)
    scores = mock_score_function(logits)
    excepted_sets = (scores <= q_hat).int()
    assert torch.equal(pred_sets, excepted_sets)


def test_invalid_predict_model(mock_score_function, mock_dataset):
    predictor = SplitPredictor(mock_score_function, None)
    with pytest.raises(ValueError, match="Model is not defined"):
        predictor.predict(mock_dataset.x)


def test_q_hat_value_error(mock_model, mock_score_function):
    """Test that a ValueError is raised when self.q_hat is None and q_hat is not provided."""
    predictor = SplitPredictor(score_function=mock_score_function, model=mock_model)

    # Ensure self.q_hat is None
    predictor.q_hat = None

    logits = torch.rand(10, 5)

    with pytest.raises(ValueError, match="Ensure self.q_hat is not None. Please perform calibration first."):
        predictor.predict_with_logits(logits)


@pytest.mark.parametrize("q_hat", [0.5, 0.7])
def test_evaluate(predictor, mock_score_function, mock_model, mock_dataset, q_hat):
    cal_dataloader = torch.utils.data.DataLoader(mock_dataset, batch_size=40)
    predictor.q_hat = q_hat
    results = predictor.evaluate(cal_dataloader)

    logits = mock_model(mock_dataset.x)
    scores = mock_score_function(logits)
    excepted_sets = (scores <= q_hat).int()

    metrics = Metrics()
    assert len(results) == 2
    assert results['coverage_rate'] == metrics('coverage_rate')(excepted_sets, mock_dataset.labels)
    assert results['average_size'] == metrics('average_size')(excepted_sets)


def test_predict_p(predictor, mock_score_function, mock_model, mock_dataset):
    with pytest.raises(ValueError):
        tmp_predictor = SplitPredictor(mock_score_function)
        tmp_predictor.predict_p(mock_dataset.x)

    logits = mock_model(mock_dataset.x)
    logits = predictor._logits_transformation(logits)

    predictor.calculate_threshold(logits, mock_dataset.labels)
    p_values = predictor.predict_p(mock_dataset.x)
    
    assert p_values.shape == (mock_dataset.x.shape[0], logits.shape[1])
    
    p_values = predictor.predict_p(mock_dataset.x, mock_dataset.labels)
    assert p_values.shape == (mock_dataset.x.shape[0], )

    p_values = predictor.predict_p(mock_dataset.x, mock_dataset.labels, True)
    assert p_values.shape == (mock_dataset.x.shape[0], )



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

"""Implementation of different scoring variants."""

import torch

import torch.nn.functional as F


def geometric(p, mask=None):
    """Score of a set is based on a geometric distribution approximation:

    p(\exists y \in C : L(y) = 0) = 1 - \prod 1 - p(A(y_i) = 1)

    This is consistent with -sum \log (1 - p(A(y_i) = 1)).

    Args:
        p: Matrix of size [num_examples, max_size].
           Each entry of p approximates p(A(y_ij) = 1).

    Returns:
        Log geometric scores.
    """
    if mask is not None:
        p = p * mask
    p = torch.maximum(1 - p, torch.tensor(1e-8))
    return -torch.cumsum(torch.log(p), dim=-1)


def marginal(p, mask=None):
    """Similar to geometric, but with p(y_k is the only y with A(y) = 1)."""
    if mask is not None:
        p = p * mask
    p = torch.maximum(1 - p, torch.tensor(1e-8))
    shifted = F.pad(p, (1, 0), mode='constant', value=1.0)[:, :-1]
    return -torch.log(1 - p) - torch.cumsum(torch.log(shifted), dim=-1)


def first_k(X, mask=None):
    """Scores are equal to the number of draws."""
    num_examples, max_generations = X.shape
    scores = torch.ones((num_examples, max_generations))
    if mask is not None:
        scores = scores * mask
    return torch.cumsum(scores, dim=-1)


def first_k_no_mask(X, mask=None):
    """Scores are equal to the number of draws."""
    del mask
    num_examples, max_generations = X.shape
    scores = torch.ones((num_examples, max_generations))
    return torch.cumsum(scores, axis=-1)


def max(X, mask=None):
    if mask is not None:
        X = X * mask
    cumulative_max, _ = torch.cummax(X, dim=-1)
    return cumulative_max


def sum(X, mask=None):
    if mask is not None:
        X = X * mask
    return torch.cumsum(X, dim=-1)



import pytest
import torch
import torch.nn as nn
from torch.utils.data import TensorDataset, DataLoader
from tqdm import tqdm
from torchcp.regression.score.abs import ABS
from torchcp.regression.score.norabs import NorABS, DifficultyEstimator, TorchMinMaxScaler

import torch
from torch.utils.data import TensorDataset, DataLoader

from torchcp.regression.predictor import SplitPredictor
from torchcp.regression.utils import build_regression_model
from torchcp.regression.score.norabs import NorABS


def test_splitpredictor_with_norabs():
    X = torch.randn(40, 3)
    y = torch.randn(40)
    train_dataset = TensorDataset(X[:20], y[:20])
    cal_dataset = TensorDataset(X[20:30], y[20:30])
    test_dataset = TensorDataset(X[30:], y[30:])

    train_loader = DataLoader(train_dataset, batch_size=8, shuffle=True)
    cal_loader = DataLoader(cal_dataset, batch_size=8)
    test_loader = DataLoader(test_dataset, batch_size=4)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    model = build_regression_model("GaussianRegressionModel")(X.shape[1], 16, 0.1)

    score_function = NorABS(data_loader=train_loader, estimate_type="variance", k=5, scalar=True, beta=0.01, device=device)
    predictor = SplitPredictor(score_function=score_function, model=model, alpha=0.1, device=device)

    predictor.train(train_loader, epochs=3, lr=0.01, device=device, verbose=True)

    predictor.calibrate(cal_loader)
    
    for i in tqdm(range(len(test_dataset)), desc="Online Evaluation"):
        x_test, y_test = test_dataset[i]
        x_test = x_test.to(device)
        pred_interval = predictor.predict(x_test.unsqueeze(0))[0][0]

        assert pred_interval.shape == (2,), "Prediction interval should be of shape (2,)"
        assert pred_interval[0] <= pred_interval[1], "Lower bound should not exceed upper bound"
        break 

# -------------------------------
# Fixtures
# -------------------------------
@pytest.fixture
def dummy_data():
    X = torch.randn(50, 4)
    y = torch.randn(50)
    dataset = TensorDataset(X, y)
    loader = DataLoader(dataset, batch_size=10)
    return loader


@pytest.fixture
def simple_model():
    """A trivial regression model for calibration tests."""
    class SimpleModel(nn.Module):
        def __init__(self, in_dim=4):
            super().__init__()
            self.fc = nn.Linear(in_dim, 2)  # mean + variance

        def forward(self, x):
            out = self.fc(x)
            return torch.cat([out[:, :1], torch.abs(out[:, 1:]) + 1e-6], dim=1)

    return SimpleModel()


# -------------------------------
# TorchMinMaxScaler Tests
# -------------------------------
def test_minmax_scaler_fit_transform_and_clip():
    data = torch.tensor([[1.0, 2.0], [3.0, 4.0]])
    scaler = TorchMinMaxScaler(clip=True)
    out = scaler.fit_transform(data)
    assert torch.all((out >= 0) & (out <= 1)), "Scaler should produce values in [0, 1]"


def test_minmax_scaler_no_fit_raises():
    scaler = TorchMinMaxScaler()
    with pytest.raises(RuntimeError):
        _ = scaler.transform(torch.randn(2, 2))


def test_minmax_scaler_constant_feature():
    data = torch.ones(5, 2)
    scaler = TorchMinMaxScaler().fit(data)
    out = scaler.transform(data)
    assert torch.allclose(out, torch.zeros_like(out)), "Constant feature should normalize to 0"


# -------------------------------
# DifficultyEstimator Tests
# -------------------------------
def test_invalid_estimator_type(dummy_data):
    with pytest.raises(ValueError):
        DifficultyEstimator(dummy_data, estimator_type="invalid")


def test_function_mode_without_callable(dummy_data):
    with pytest.raises(ValueError):
        DifficultyEstimator(dummy_data, estimator_type="function")


def test_variance_mode_with_invalid_predicts(dummy_data, simple_model):
    est = DifficultyEstimator(dummy_data, estimator_type="variance")
    est.calibrate(simple_model)
    with pytest.raises(ValueError):
        est._compute_difficulty(torch.randn(2, 2), torch.randn(2, 3))


def test_function_mode_with_invalid_inputs(dummy_data, simple_model):
    est = DifficultyEstimator(dummy_data, estimator_type="function", custom_function=lambda x, y: y[:, 1])
    est.calibrate(simple_model)
    with pytest.raises(ValueError):
        est._compute_difficulty(None, None)


def test_knn_modes(dummy_data, simple_model):
    for mode in ["knn_distance", "knn_label", "knn_residual"]:
        est = DifficultyEstimator(dummy_data, estimator_type=mode, k=3)
        est.calibrate(simple_model)
        scores = est.apply(torch.randn(2, 4))
        assert scores.shape[0] == 2


def test_apply_without_calibrate(dummy_data):
    est = DifficultyEstimator(dummy_data, estimator_type="variance")
    with pytest.raises(RuntimeError):
        est.apply(torch.randn(2, 4), torch.randn(2, 2))


# -------------------------------
# NorABS Tests
# -------------------------------
def test_norabs_call_and_generate_intervals(dummy_data, simple_model):
    norabs = NorABS(dummy_data, estimate_type="variance")
    norabs.calibrate(simple_model)

def test_norabs_call_and_calibration(dummy_data, simple_model):
    norabs = NorABS(dummy_data, estimate_type="variance")
    model = simple_model
    x = torch.randn(2, 1)
    predicts = torch.randn(2, 2)
    y_truth = torch.randn(2)
    scores = norabs(predicts, y_truth, x_batch=x, model=model)
    assert scores.shape == (2, 1)


def test_norabs_call_without_model_raises(dummy_data):
    norabs = NorABS(dummy_data, estimate_type="variance")
    with pytest.raises(ValueError):
        norabs(torch.randn(2, 2), torch.randn(2))


def test_norabs_call_without_model(dummy_data):
    norabs = NorABS(dummy_data, estimate_type="variance")
    predicts = torch.randn(3, 2)
    y_truth = torch.randn(3)
    with pytest.raises(ValueError):
        _ = norabs(predicts, y_truth)


def test_norabs_train_returns_model(dummy_data):
    norabs = NorABS(dummy_data, estimate_type="variance")
    model = norabs.train(dummy_data, epochs=1, verbose=False)
    sample_input, _ = next(iter(dummy_data))
    out = model(sample_input)
    assert out.shape[1] == 2



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import pytest
import torch
from torch_geometric.data import Data

from torchcp.classification.score import LAC
from torchcp.graph.score import DAPS
from torchcp.graph.score.base import BaseScore


@pytest.fixture
def graph_data():
    x = torch.tensor([
        [1.0, 2.0],
        [3.0, 4.0],
        [5.0, 6.0]
    ], dtype=torch.float32)

    edge_index = torch.tensor([
        [0, 1, 1, 2],
        [1, 0, 2, 1]
    ], dtype=torch.long)

    edge_weight = torch.ones(edge_index.shape[1])

    y = torch.tensor([0, 1, 0], dtype=torch.long)

    data = Data(x=x, edge_index=edge_index, edge_weight=edge_weight, y=y)

    return data


@pytest.fixture
def base_score_function():
    return LAC(score_type="softmax")


def test_base_graph_scores(graph_data, base_score_function):
    class TestScores(BaseScore):
        def __init__(self, graph_data, base_score_function):
            super().__init__(graph_data, base_score_function)

    score_function = TestScores(graph_data, base_score_function)
    with pytest.raises(NotImplementedError):
        score_function(None)


def test_daps_initialization(graph_data, base_score_function):
    daps = DAPS(graph_data, base_score_function, neigh_coef=0.7)
    assert daps._n_vertices == graph_data.num_nodes
    assert daps._device == graph_data.edge_index.device

    def are_sparse_tensors_equal(tensor1, tensor2):
        if tensor1.shape != tensor2.shape:
            return False
        if tensor1.device != tensor2.device:
            return False
        if not torch.equal(tensor1.coalesce().indices(), tensor2.coalesce().indices()):
            return False
        if not torch.equal(tensor1.coalesce().values(), tensor2.coalesce().values()):
            return False
        return True

    edge_weight = torch.ones(graph_data.edge_index.shape[1])
    adj = torch.sparse_coo_tensor(
        graph_data.edge_index,
        edge_weight,
        (graph_data.num_nodes, graph_data.num_nodes))
    assert are_sparse_tensors_equal(daps._adj, adj)

    edge_weight = graph_data.edge_weight
    adj = torch.sparse_coo_tensor(
        graph_data.edge_index,
        edge_weight,
        (graph_data.num_nodes, graph_data.num_nodes))
    assert are_sparse_tensors_equal(daps._adj, adj)

    degs = torch.matmul(adj, torch.ones((adj.shape[0])))
    assert daps._degs.equal(degs)


@pytest.mark.parametrize("neigh_coef", [-0.1, 1.1, -5, 1.5])
def test_invalid_neigh_coef(graph_data, base_score_function, neigh_coef):
    with pytest.raises(ValueError, match="The parameter 'neigh_coef' must be a value between 0 and 1."):
        DAPS(graph_data, base_score_function, neigh_coef)


def test_daps_call_without_labels(graph_data, base_score_function):
    daps = DAPS(graph_data, base_score_function, neigh_coef=0.5)
    logits = torch.tensor([
        [1.0, 0.5],
        [0.2, 0.8],
        [0.4, 0.6]
    ], dtype=torch.float32)

    base_scores = base_score_function(logits)

    diffusion_scores = torch.tensor([
        [base_scores[1, 0], base_scores[1, 1]],
        [(base_scores[0, 0] + base_scores[2, 0]) / 2, (base_scores[0, 1] + base_scores[2, 1]) / 2],
        [base_scores[1, 0], base_scores[1, 1]]
    ], dtype=torch.float32)

    expected_scores = 0.5 * diffusion_scores + 0.5 * base_scores

    scores = daps(logits)
    assert torch.allclose(scores, expected_scores, atol=1e-5)


def test_daps_call_with_labels(graph_data, base_score_function):
    daps = DAPS(graph_data, base_score_function, neigh_coef=0.5)

    logits = torch.tensor([
        [1.0, 0.5],
        [0.2, 0.8],
        [0.4, 0.6]
    ], dtype=torch.float32)

    labels = torch.tensor([0, 1, 0], dtype=torch.long)

    scores = daps(logits, labels)
    expected_scores = daps(logits)
    assert torch.allclose(scores, expected_scores[torch.arange(3), labels], atol=1e-5)



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import os

import pytest
import torch
from torch import nn
from torch.utils.data import DataLoader, TensorDataset

from torchcp.classification.trainer import ConfTrTrainer



# Mock model for testing
class MockModel(nn.Module):
    def __init__(self):
        super().__init__()
        self.linear = nn.Linear(10, 3)
        
    def forward(self, x):
        return self.linear(x)

# Fixtures
@pytest.fixture
def mock_model():
    return MockModel()

@pytest.fixture
def mock_data():
    x = torch.randn(1000, 10)
    y = torch.randint(0, 3, (1000,))
    dataset = TensorDataset(x, y)
    return DataLoader(dataset, batch_size=128, shuffle=True)

@pytest.fixture
def confts_trainer(mock_model):
    return ConfTrTrainer(
        alpha=0.1,
        model=mock_model
    )

# Tests
class TestConfTrTrainer:
    def test_initialization(self, mock_model):
        trainer = ConfTrTrainer(
            alpha=0.1,
            model=mock_model
        )
        
    def test_invalid_params(self, mock_model):
            
        # Test invalid alpha
        with pytest.raises(ValueError):
            ConfTrTrainer(
                alpha=-0.1,
                model=mock_model
            )


# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import pytest

from torchcp.utils.registry import Registry


def test_registry_init():
    """Test registry initialization"""
    registry = Registry("test")
    assert registry._name == "test"
    assert registry._obj_map == {}


def test_registry_basic_registration():
    """Test basic object registration"""
    registry = Registry("test")

    class DummyClass:
        pass

    # Test function call registration
    registry.register(DummyClass)
    assert "DummyClass" in registry.registered_names()
    assert registry.get("DummyClass") == DummyClass


def test_registry_decorator():
    """Test decorator registration"""
    registry = Registry("test")

    @registry.register()
    class TestClass:
        pass

    assert "TestClass" in registry.registered_names()
    assert registry.get("TestClass") == TestClass


def test_duplicate_registration():
    """Test handling of duplicate registration"""
    registry = Registry("test")

    class DummyClass:
        pass

    registry.register(DummyClass)

    # Try to register the same class again
    with pytest.raises(KeyError, match="was already registered"):
        registry.register(DummyClass)

    # Test force registration
    registry.register(DummyClass, force=True)
    assert registry.get("DummyClass") == DummyClass


def test_get_nonexistent():
    """Test getting non-existent object"""
    registry = Registry("test")

    with pytest.raises(KeyError, match="does not exist"):
        registry.get("NonexistentClass")


def test_registered_names():
    """Test listing registered names"""
    registry = Registry("test")

    # Register multiple classes
    class Class1:
        pass

    class Class2:
        pass

    registry.register(Class1)
    registry.register(Class2)

    names = registry.registered_names()
    assert isinstance(names, list)
    assert "Class1" in names
    assert "Class2" in names
    assert len(names) == 2



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import torch
from typing import List

from torchcp.regression.predictor.aci import ACIPredictor

class AgACIPredictor(ACIPredictor):
    """
    Online Expert Aggregation Adaptive Conformal Inference.
    
    A parameter-free method that adaptively builds upon ACI based on online expert aggregation.
    
    Args:
        score_function (torchcp.regression.scores): A class that implements the score function.
        model (torch.nn.Module): A PyTorch model capable of outputting quantile values.
            The model should be an initialization model that has not been trained.
        gamma_list (List[float]): A list of step size parameters for adaptive adjustment of alpha at each step.
            Each element in the list corresponds to a different expert. Must contain values greater than 0.
        aggregation_function (str or callable, optional): The function used to aggregate predictions from experts.
            Can be either 'mean' (average), 'median', or a custom callable function. Defaults to 'mean'.
        threshold (List[float], optional): A list containing the lower and upper thresholds for clipping expert predictions.
            Defaults to [-99999, 99999] (effectively no clipping).
        alpha (float, optional): The significance level. Default is 0.1.
        device (torch.device, optional): The device on which the model is located. Default is None.
        
    Reference:  
        Paper: Adaptive Conformal Predictions for Time Series (Zaffran et al., 2022)
        Link: https://proceedings.mlr.press/v162/zaffran22a.html
        Github: https://github.com/mzaffran/AdaptiveConformalPredictionsTimeSeries
        
    """
    
    def __init__(self, score_function, model, gamma_list: List, aggregation_function='mean', threshold: List=[-99999, 99999], alpha=0.1, device=None):
        super().__init__(score_function, model, None, alpha, device)
        if aggregation_function not in ['mean', 'median'] and not callable(aggregation_function):
            raise ValueError(
                "aggregation_function must be either 'mean', 'median', or a callable function."
            )
        if not isinstance(gamma_list, list):
            raise ValueError(f"gamma_list must be a list, but got {type(gamma_list).__name__}.")
        if not isinstance(threshold, list) or len(threshold) != 2:
            raise ValueError(f"threshold must be a list with exactly 2 elements, but got {threshold}.")
        

        self.gamma_list = gamma_list
        self.alpha_t = None
        self.model_backbone = model
        self.train_indicate = False
        
        if aggregation_function == 'mean':
            self.aggregation_function = torch.mean
        elif aggregation_function == 'median':
            self.aggregation_function = lambda x, dim: torch.median(x, dim=dim)[0]
        else:
            self.aggregation_function = aggregation_function
            
        self.lower_threshold, self.upper_threshold = threshold[0], threshold[1]
        

    def generate_aci_intervals(self, x_batch, x_lookback, y_lookback, pred_interval_lookback, predicts_batch):
        """
        Generate Aggregation Adaptive Conformal Intervals (AgACI) and compute the weighted confidence interval.

        Args:
        - x_batch: Current batch input data
        - x_lookback: Past input data
        - y_lookback: Past ground truth labels (N,)
        - pred_interval_lookback: Past prediction intervals (N, 2) [lower_bound, upper_bound]
        - predicts_batch: Model predictions for the current batch

        Returns:
        - weighted_intervals: The weighted confidence intervals
        """
        err_t = self.calculate_err_rate(x_batch, y_lookback, pred_interval_lookback, weight=False)
        scores = self.calculate_score(self._model(x_lookback).float(), y_lookback)

        intervals_list = []
        weight_list = []
        y_lookback = y_lookback.unsqueeze(1)
        for gamma in self.gamma_list:
            # Compute the adaptive alpha_t
            alpha_t = max(1 / (scores.shape[0] + 1), min(0.9999, self.alpha + gamma * (self.alpha - err_t)))
            q_hat = self._calculate_conformal_value(scores, alpha_t)
            pred_intervals = self.generate_intervals(predicts_batch, q_hat)
            
            # Ensure pred_intervals are within a valid range
            pred_intervals[:, :, 0] = torch.where(
                torch.isfinite(pred_intervals[:, :, 0]), pred_intervals[:, :, 0], self.lower_threshold
            )
            pred_intervals[:, :, 1] = torch.where(
                torch.isfinite(pred_intervals[:, :, 1]), pred_intervals[:, :, 1], self.upper_threshold
            )

            intervals_list.append(pred_intervals)

            # Compute the Pinball Loss
            quantiles = torch.tensor([self.alpha / 2, 1 - self.alpha / 2], device=y_lookback.device)
            lower_bound, upper_bound = pred_interval_lookback[:, :, 0], pred_interval_lookback[:, :, 1]

            loss_lower = torch.maximum(quantiles[0] * (y_lookback - lower_bound),
                                    (quantiles[0] - 1) * (y_lookback - lower_bound)).detach()
            loss_upper = torch.maximum(quantiles[1] * (y_lookback - upper_bound),
                                    (quantiles[1] - 1) * (y_lookback - upper_bound)).detach()

            # Compute weighted loss
            # breakpoint()
            weight = torch.tensor([[self.aggregation_function(loss_lower, dim=0).tolist(), self.aggregation_function(loss_upper, dim=0).tolist()]],
                                device=y_lookback.device) 
            weight_list.append(weight)
        # breakpoint()
        # Compute the weighted confidence interval
        stacked_intervals = torch.stack(intervals_list)   # (num_experts, x_batch_size, 1, 2)
        stacked_weights = torch.stack(weight_list).permute(0, 3, 1, 2)  
        
        # Normalize weights and ensure the denominator is nonzero
        weight_sum = torch.sum(stacked_weights, dim=0, keepdim=True)
        weight_sum = torch.where(weight_sum == 0, torch.tensor(1.0, device=weight_sum.device), weight_sum)  # Avoid division by zero
        weighted_intervals = torch.sum(stacked_intervals * stacked_weights, dim=0) / weight_sum
        # breakpoint()
        return weighted_intervals[0]


# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import torch.nn as nn


class BaseLoss(nn.Module):
    """
    Base class for conformal loss functions.

    Args:
        predictor (torchcp.classification.Predictor): Predictor instance that defines
            the scoring mechanism for conformal prediction.
    """

    def __init__(self, predictor):
        super(BaseLoss, self).__init__()
        self.predictor = predictor

    def forward(self, predictions, targets):
        raise NotImplementedError



# Copyright (c) 2023-present, SUSTech-ML.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import pytest
import torch

from torchcp.regression.predictor import ACIPredictor
from torchcp.regression.score import CQR
from torchcp.regression.utils import build_regression_model


@pytest.fixture
def mock_model():
    return build_regression_model("NonLinearNet")(5, 2, 64, 0.5)


@pytest.fixture
def mock_score_function():
    return CQR()


def test_aci_predictor_workflow(mock_data, mock_model, mock_score_function):
    train_dataloader, _, test_dataloader = mock_data

    with pytest.raises(ValueError, match="gamma must be greater than 0."):
        ACIPredictor(mock_score_function, mock_model, gamma=0)

    alpha = 0.1
    # Initialize ACIPredictor
    aci_predictor = ACIPredictor(
        score_function=mock_score_function,
        model=mock_model,
        gamma=0.1,
        alpha=alpha
    )

    # Test train method
    aci_predictor.train(train_dataloader)
    assert aci_predictor.alpha == alpha, "Alpha should be set correctly during training."
    assert aci_predictor.alpha_t == alpha, "Adaptive alpha_t should start with initial alpha."

    # Test predict method with x_batch only
    x_batch_1, y_batch_1 = next(iter(test_dataloader))
    prediction_intervals_1 = aci_predictor.predict(x_batch_1)
    assert prediction_intervals_1 is not None, "Prediction intervals should not be None."
    assert prediction_intervals_1.shape[0] == x_batch_1.shape[0], "Prediction intervals should match batch size."
    # Test exception for missing arguments
    with pytest.raises(ValueError):
        aci_predictor.predict(x_batch_1, y_lookback=torch.rand(10), pred_interval_lookback=None)
    with pytest.warns(UserWarning):
        prediction_intervals_1 = aci_predictor.predict(x_batch_1, train=True, update_alpha=True)
    assert prediction_intervals_1 is not None, "Prediction intervals should not be None."
    assert prediction_intervals_1.shape[0] == x_batch_1.shape[0], "Prediction intervals should match batch size."
    # Test predict method with x_batch, x_lookback, y_lookback
    x_batch_2, y_batch_2 = next(iter(test_dataloader))
    prediction_intervals_2 = aci_predictor.predict(x_batch=x_batch_2, x_lookback=x_batch_1, y_lookback=y_batch_1)
    assert prediction_intervals_2 is not None, "Prediction intervals should not be None."
    assert prediction_intervals_2.shape[0] == x_batch_2.shape[0], "Prediction intervals should match batch size."
    # Test predict method with x_batch, x_lookback, y_lookback, pred_interval_lookack
    x_batch_3, y_batch_3 = next(iter(test_dataloader))
    prediction_intervals_3 = aci_predictor.predict(x_batch=x_batch_3, x_lookback=x_batch_2, y_lookback=y_batch_2,
                                                   pred_interval_lookback=prediction_intervals_2)
    assert prediction_intervals_3 is not None, "Prediction intervals should not be None."
    assert prediction_intervals_3.shape[0] == x_batch_3.shape[0], "Prediction intervals should match batch size."

    # Test evaluate method
    eval_results = aci_predictor.evaluate(test_dataloader)
    assert eval_results["coverage_rate"] > 0, "Average coverage rate should be greater than 0."
    assert eval_results["average_size"] > 0, "Average interval size should be greater than 0."
    # Test evaluate method when lookback is too high
    with pytest.raises(ValueError):
        eval_results = aci_predictor.evaluate(test_dataloader, lookback=10000)
    # Test evaluate method with other arguments
    aci_predictor.evaluate(test_dataloader, retrain_gap=0, update_alpha_gap=0)
    aci_predictor.evaluate(test_dataloader, retrain_gap=2, update_alpha_gap=3)


def test_aci_predictor_wrong_workflow(mock_data, mock_model, mock_score_function):
    train_dataloader, _, test_dataloader = mock_data

    with pytest.raises(ValueError, match="gamma must be greater than 0."):
        ACIPredictor(mock_score_function, mock_model, gamma=0)

    aci_predictor = ACIPredictor(
        score_function=mock_score_function,
        model=mock_model,
        gamma=0.1
    )

    x_batch, y_batch = next(iter(test_dataloader))
    # predict should used after train
    with pytest.raises(ValueError):
        prediction_intervals = aci_predictor.predict(x_batch)


@pytest.mark.parametrize("device", ["cpu", "cuda"])
def test_device_support(mock_data, mock_model, mock_score_function, device):
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA is not available.")

    train_dataloader, _, test_dataloader = mock_data
    aci_predictor = ACIPredictor(mock_score_function, mock_model.to(device), gamma=0.1)

    aci_predictor.train(train_dataloader)
    x_batch, _ = next(iter(test_dataloader))
    x_batch = x_batch.to(device)

    prediction_intervals = aci_predictor.predict(x_batch)
    assert prediction_intervals.device.type == device, "Prediction intervals should match the specified device."
